from matrx.agents.capabilities.capability import SenseCapability
from matrx.actions.action import Action
from matrx.objects.env_object import EnvObject


class AgentBody(EnvObject):

    def __init__(self, location, possible_actions, sense_capability, class_callable,
                 callback_agent_get_action, callback_agent_set_action_result, callback_agent_observe,
                 callback_agent_get_messages, callback_agent_set_messages, callback_agent_initialize,
                 callback_agent_log, callback_create_context_menu_for_other, callback_create_context_menu_for_self,
                 visualize_size, visualize_shape, visualize_colour, visualize_depth, visualize_opacity,
                 visualize_when_busy, is_traversable, team, name, is_movable,
                 is_human_agent, customizable_properties,
                 **custom_properties):
        """
        This class is a representation of an agent's body in the GridWorld.

        It is used as a measure to keep the AgentBrain code and Environment code separate. This AgentBody is used by
        the environment to update the GUI, perform actions, and update properties. It is kept in sync with the Agent's
        brain every iteration.

        It inherits from EnvObject which allows you set any custom properties you want. In addition it also has all the
        mandatory properties of an EnvObject plus a few extra. Which is the team name the agent is part of (if any) and
        what the Agent's body is carrying.

        In addition the Agent's body keeps a set of callbacks to methods inside the Agent. This forms the connection
        between the GridWorld (that calls them) and the Agent (that defined them).

        :param location: List or tuple of length two. Mandatory. The location of the Agent's body in the grid world.
        :param possible_actions: The list of Action class names this agent may be able to perform. This allows you to
        create agents that can only perform a couple of the available actions.
        :param sense_capability: The SenseCapability object.
        :param class_callable: The Agent class; in other words, the class of the agent's brain. This is stored here so
        that the Visualizer (which visualizes an agent based on this Agent's body object) and agents knows what kind of agent
        it is. Allows you to visualize certain agent types in a certain way.

        :param callback_agent_get_action: The callback function as defined by the Agent instance of which this is an
        Agent's body of. It is called each tick by the GridWorld. As such the GridWorld determines when the Agent can
        perform an action by calling this function which is stored in the Agent's Agent's body.
        :param callback_agent_set_action_result: Same as the callback_get_action but is used by GridWorld to set the
        ActionResult object in the Agent after performing the action. This allows the Agent to know how its planned
        action went.
        :param callback_agent_observe: Similar to callback_agent_get_action, is used by GridWorld to obtain the
        processed state dictionary of the Agent. As the GridWorld does not know exactly what the Agent is allowed to
        see or not, the 'observe' preprocesses the given state further. But to accurately visualize what the agent sees
        we have to obtain that pre-processed state, which is done through this callback.
        :param callback_agent_initialize: Call the initialize function in an Agent's Brain. Is done at the initialize
        of a GridWorld.

        :param callback_agent_get_messages: A callback function that allows the GridWorld to obtain the agent's messages
        that need to be send to different agents.
        :param callback_agent_set_messages: A callback function that allows the GridWorld to set any message send by
        some agent to a list of received messages in an agent.

        :param callback_create_context_menu_for_other: A callback function that allows the gridworld or API to call the
        subsequent agent function that generates the menu options of an agent for a context menu opened by a user not
        controlling that specific agent.
        :param callback_create_context_menu_for_self: A callback function that allows the gridworld or API to call the
        subsequent agent function that generates the menu options for a context menu opened by the user controlling
        the current human agent. If this is not a human agent, it is set to None.


        :param name: String Defaults to "Agent". The name of the agent, does not need to be unique.
        :param is_human_agent: Boolean. Defaults to False. Boolean to signal that the agent represented by this Agent's
        body is a human controlled agent.
        :param customizable_properties: List. Optional, default obtained from defaults.py. The list of attribute names
        that can be customized by other objects (including Agent's body and as an extension any Agent).
        :param is_traversable: Boolean. Optional, default obtained from defaults.py. Signals whether other objects can
        be placed on top of this object.
        :param carried_by: List. Optional, default obtained from defaults.py. A list of who is carrying this object.
        :param team: The team name the agent is part of. Defaults to the team name similar to the Agent's body unique
        ID, as such denoting that by default each Agent's body belongs to its own team and as an extension so does its
        "brain" the Agent.

        :param visualize_size: Float. Optional, default obtained from defaults.py. A visualization property used by
        the Visualizer. Denotes the size of the object, its unit is a single grid square in the visualization (e.g. a
        value of 0.5 is half of a square, object is in the center, a value of 2 is twice the square's size centered on
        its location.)
        :param visualize_shape: Int. Optional, default obtained from defaults.py. A visualization property used by the
        Visualizer. Denotes the shape of the object in the visualization.
        :param visualize_colour: Hexcode string. Optional, default obtained from defaults.py. A visualization property
        used by the Visualizer. Denotes the
        :param visualize_depth: Integer. Optional, default obtained from defaults.py. A visualization property that
        is used by the Visualizer to draw objects in layers.
        :param visualize_opacity: Integer. Opacity of object. Between 0.0 and 1.0.
        :param visualize_when_busy: Boolean. Whether to show a loading icon when the agent is busy (performing an action).
        :param **custom_properties: Optional. Any other keyword arguments. All these are treated as custom attributes.
        For example the property 'heat'=2.4 of an EnvObject representing a fire.
        """

        # A list of EnvObjects or any class that inherits from it. Denotes all objects the Agent's body is currently
        # carrying. Note that these objects do not exist on the WorldGrid anymore, so removing them in this list deletes
        # them permanently.
        self.is_carrying = []  # list of EnvObjects that this object carries

        # The property that signals whether the agent this Agent's body represents is a human agent
        self.is_human_agent = is_human_agent

        # Save the other attributes the GridWorld expects an Agent's body to have access to an Agent's brain
        self.get_action_func = callback_agent_get_action
        self.set_action_result_func = callback_agent_set_action_result
        self.filter_observations = callback_agent_observe
        self.get_messages_func = callback_agent_get_messages
        self.set_messages_func = callback_agent_set_messages
        self.get_log_data = callback_agent_log
        self.brain_initialize_func = callback_agent_initialize
        self.create_context_menu_for_other_func = callback_create_context_menu_for_other
        self.create_context_menu_for_self_func = callback_create_context_menu_for_self

        # Set all mandatory properties
        self.is_traversable = is_traversable
        self.sense_capability = sense_capability
        self.action_set = possible_actions
        self.is_movable = is_movable

        # Set visualization properties
        self.visualize_depth = visualize_depth
        self.visualize_colour = visualize_colour
        self.visualize_shape = visualize_shape
        self.visualize_size = visualize_size
        self.visualize_opacity = visualize_opacity
        self.visualize_when_busy = visualize_when_busy

        # Parse the action_set property if set to the wildcard "*" denoting all actions
        if self.action_set == "*":
            self.action_set = list(_get_all_classes(Action, omit_super_class=True).keys())

        # Defines an agent is blocked by an action which takes multiple time steps. Is updated based on the speed with
        # which an agent can perform actions.
        self.__is_blocked = False

        # Place holders for action information
        self.__current_action = None
        self.__current_action_args = None

        # Denotes the last action performed by the agent, at what tick and how long it must take. Set to -infinite, such
        # that the agent is not deemed 'busy' at tick 0. Edit: changed from -inf to 1000000, to be JSON serializable
        self.__last_action_duration_data = {"duration_in_ticks": -1000000, "tick": -1000000, "action_name": None,
                                            "action_result": None}

        # We set a placeholder for the 'team' property so that it can be found in self.properties
        self.team = ""

        # Call the super constructor (we do this here because then we have access to all of EnvObject, including a
        # unique id
        super().__init__(location, name, customizable_properties=customizable_properties, is_traversable=is_traversable,
                         class_callable=class_callable,
                         visualize_size=visualize_size, visualize_shape=visualize_shape,
                         visualize_colour=visualize_colour, visualize_depth=visualize_depth,
                         visualize_opacity=visualize_opacity,
                         **custom_properties)

        # the GUI cannot differentiate capital letters (in the url), so make the agent ID lowercase
        self.obj_id = self.obj_id.lower()

        # If there was no team name given, the Agent's body (and as an extension its Agent's brain) is part of its own
        # team which is simply its object id + "_team". For this we need the object id, which was made in the EnvObject
        # constructor, that is why we call this AFTER calling that.
        if team is None:
            self.team = self.obj_id + "_team"
        self.change_property("team", self.team)

    def _set_agent_busy(self, curr_tick, action_duration):
        """
        specify the duration of the action in ticks currently being executed by the
        agent, and its starting tick
        """
        self.__last_action_duration_data = {"duration_in_ticks": action_duration, "tick": curr_tick}

    def _check_agent_busy(self, curr_tick):
        """
        check if the agent is done with executing the action
        """
        self.__is_blocked = curr_tick <= (self.current_action_tick_started + self.current_action_duration_in_ticks)

        return self.__is_blocked

    def _at_last_action_duration_tick(self, curr_tick):
        """ Returns True if this agent is at its last tick of the action's duration."""
        is_last_tick = curr_tick == (self.current_action_tick_started + self.current_action_duration_in_ticks)
        return is_last_tick

    def _get_duration_action(self):
        """ Returns the action we are waiting for 'self.current_action_duration_in_ticks', gets called in the GridWorld
        when we are at the last tick on which we should wait (see self._at_last_action_duration_tick)."""
        action_name = self.current_action
        action_kwargs = self.current_action_args

        return action_name, action_kwargs

    def _set_current_action(self, action_name, action_args):
        """
        Sets the current action of the agent. Since the GridWorld performs the mutate of an action first, and then waits
        for the duration to pass, we also have the result available.
        """
        self.__current_action = action_name
        self.__current_action_args = action_args

    def _set_agent_changed_properties(self, props: dict):
        """
        The Agent has possibly changed some of its properties during its OODA loop. Here the agent properties are also
        updated in the Agent's body, if it is allowed to change them as defined in 'customizable_properties' list.
        """
        # get all agent properties of this Agent's body in one dictionary
        body_properties = self.properties

        # check for each property if it has been changed by the agent, and if we need
        # to update our local copy (here in Agent's body) of the agent properties to match that
        for prop in props.keys():
            if prop not in body_properties.keys():
                raise Exception(f"Agent {self.obj_id} tried to remove the property {prop}, which is not allowed.")

            # check if the property has changed, and skip if not the case
            if props[prop] == body_properties[prop]:
                continue

            # if the agent has changed the property, check if the agent has permission to do so
            if prop not in self.customizable_properties:
                raise Exception(f"Agent {self.obj_id} tried to change a non-writable property: {prop}.")

            # The agent changed the property and the agent had permission to do so
            # update special properties
            self.change_property(prop, props[prop])

    def change_property(self, property_name, property_value):
        """
        Changes the value of an existing (!) property.
        :param property_name: The name of the property.
        :param property_value:  The value of the property.
        :return: The new properties.
        """

        # We check if it is a custom property and if so change it simply in the dictionary
        if property_name in self.custom_properties.keys():
            self.custom_properties[property_name] = property_value
        else:  # else we need to check if property_name is a mandatory class attribute that is also a property
            if property_name == "is_traversable":
                assert isinstance(property_value, bool)
                self.is_traversable = property_value
            elif property_name == "name":
                assert isinstance(property_value, str)
                self.obj_name = property_value
            elif property_name == "location":
                assert isinstance(property_value, list) or isinstance(property_value, tuple)
                self.location = property_value
            elif property_name == "class_inheritance":
                assert isinstance(property_value, list)
                self.class_inheritance = property_value
            elif property_name == "visualize_size":
                assert isinstance(property_value, int)
                self.visualize_size = property_value
            elif property_name == "visualize_colour":
                assert isinstance(property_value, str)
                self.visualize_colour = property_value
            elif property_name == "visualize_opacity":
                assert isinstance(property_value, int)
                self.visualize_opacity = property_value
            elif property_name == "visualize_when_busy":
                assert isinstance(property_value, bool)
                self.visualize_when_busy = property_value
            elif property_name == "visualize_shape":
                assert isinstance(property_value, int)
                self.visualize_shape = property_value
            elif property_name == "visualize_depth":
                assert isinstance(property_value, int)
                self.visualize_depth = property_value
            elif property_name == "team":
                assert isinstance(property_value, str)
                self.team = property_value
            elif property_name == "sense_capability":
                assert isinstance(property_value, SenseCapability)
                self.sense_capability = property_value
            elif property_name == "is_human_agent":
                assert isinstance(property_value, bool)
                self.is_human_agent = property_value
            elif property_name == "action_set":
                assert isinstance(property_value, list)
                self.action_set = property_value
            elif property_name == "is_movable":
                assert isinstance(property_value, bool)
                self.is_movable = property_value
            # We deliberately ignore the current_action property, and several others such as agent_id as these can never
            # be altered as they are governed by the GridWorld

        return self.properties

    @property
    def location(self):
        """
        We override the location pythonic property here so we can override its setter.
        :return: The location tuple of the form; (x, y).
        """
        return tuple(self.__location)

    @location.setter
    def location(self, loc):
        """
        Overrides the setter of the location (pythonic) property so we can transfer also all carried objects with us
        on any location change made anywhere.
        :param loc:
        :return:
        """
        assert isinstance(loc, list) or isinstance(loc, tuple)
        assert len(loc) == 2
        # Set the location to our private location xy list
        self.__location = loc

        # Carrying action is done here
        # First we check if we even have a 'carrying' property, as the future might hold an Agent's body who
        # specifically removes this property. In that case we return.
        if 'carrying' not in self.properties.keys():
            return
        # Next we retrieve whatever it is the Agent's body is carrying (if we have a 'carrying' property at all)
        carried_objs = self.properties['carrying']
        # If we carry nothing, we are done
        if len(carried_objs) == 0:
            return
        # Otherwise we loop over all objects and adjust their location accordingly (since these are also EnvObjects,
        # their setter for location gets called, in the case we are carrying an Agent's body this setter is called
        for obj in carried_objs:
            obj.location = loc  # this requires all objects in self.properties['carrying'] to be of type EnvObject

    @property
    def properties(self):
        """
        Returns the custom properties of this object, but also any mandatory properties such as location, name,
        is_traversable and all visualization properties (those are in their own dictionary under 'visualization').

        In the case we return the properties of a class that inherits from EnvObject, we check if that class has

        :return: All mandatory and custom properties in a dictionary.
        """

        # Copy the custom properties
        properties = self.custom_properties.copy()

        # Add all mandatory properties. Make sure that these are updated if one are added to the constructor!
        properties['team'] = self.team
        properties['name'] = self.obj_name
        properties['obj_id'] = self.obj_id  # we return id as well, but this should never ever be modified!
        properties['location'] = self.location
        properties['is_movable'] = self.is_movable
        properties['action_set'] = self.action_set
        properties['carried_by'] = self.carried_by
        properties['is_human_agent'] = self.is_human_agent
        properties['is_traversable'] = self.is_traversable
        properties['class_inheritance'] = self.class_inheritance
        properties['is_blocked_by_action'] = self.is_blocked
        properties['is_carrying'] = [obj.properties for obj in self.is_carrying]
        properties['sense_capability'] = self.sense_capability.get_capabilities()
        properties['visualization'] = {
            "size": self.visualize_size,
            "shape": self.visualize_shape,
            "colour": self.visualize_colour,
            "depth": self.visualize_depth,
            "opacity": self.visualize_opacity,
            "show_busy": self.visualize_when_busy
        }

        # Add the current action and all of its data
        properties['current_action'] = self.current_action
        if self.current_action is not None:  # all None actions are 'idle' actions and have no name or result
            properties['current_action_args'] = self.current_action_args  # the action arguments
        else:
            properties['current_action_args'] = {}

        properties['current_action_duration'] = self.current_action_duration_in_ticks
        properties['current_action_started_at_tick'] = self.current_action_tick_started

        return properties

    @properties.setter
    def properties(self, property_dictionary: dict):
        """
        Here to protect the 'properties' variable. It does not do anything and should not do anything!
        """
        pass

    @property
    def current_action(self):
        return self.__current_action

    @property
    def current_action_duration_in_ticks(self):
        return self.__last_action_duration_data["duration_in_ticks"]

    @property
    def current_action_tick_started(self):
        return self.__last_action_duration_data["tick"]

    @property
    def current_action_args(self):
        return self.__current_action_args

    @property
    def is_blocked(self):
        return self.__is_blocked


def _get_all_classes(class_, omit_super_class=False):
    # Include given class or not
    if omit_super_class:
        subclasses = set()
    else:
        subclasses = {class_}

    # Go through all child classes
    work = [class_]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)

    # Create a dict out of it
    act_dict = {}
    for action_class in subclasses:
        act_dict[action_class.__name__] = action_class

    return act_dict



""" Main module for the core's Action class.
"""


class Action:
    """ Contains the effects of an action on the world.

    This class is empty and should be overridden if you want to make a
    new action that is not yet supported. You may also extend other
    actions, but all MATRX' actions inherit from this class.

    Parameters
    ----------
    duration_in_ticks : int
        The default duration of the action in ticks during which the
        GridWorld blocks the Agent performing other actions. By default
        this is 1, meaning that the action will take both the tick in which
        it was decided upon and the subsequent tick. When creating your
        own Action, you can override this default value.

    """

    def __init__(self, duration_in_ticks=0):
        # number of ticks the action takes to complete
        self.duration_in_ticks = duration_in_ticks

    def mutate(self, grid_world, agent_id, **kwargs):
        """ Method that mutates the world.

        This method is allowed to mutate a :class:`matrx.grid_world.GridWorld`
        instance according to the purposes of the
        :class:`matrx.actions.Action`. Override this method when implementing
        a new :class:`matrx.actions.action.Action`.

        Parameters
        ----------
        grid_world
            The GridWorld instance that should be mutated according to the
            Action's intended purpose.
        agent_id : string
            The unique identifier of the agent performing this action.
        **kwargs
            The set of keyword arguments provided by the agent that decided
            upon this action. When overriding this method and setting required
            arguments, the :class:`matrx.grid_world.GridWorld` will take
            responsibility in checking whether the agent provided these. Hence,
            it is not required to do so here.

        Returns
        -------
        ActionResult
            An instance of the :class:`matrx.actions.Action.ActionResult`
            class, or an inherited class thereof, which denotes whether the
            mutation actually succeeded or not, together with the reason why.
            This result is passed to the
            :class:`matrx.agents.agent_brain.AgentBrain` by the
            :class:`matrx.grid_world.GridWorld` instance so that an agent is
            capable of checking when and why an
            :class:`matrx.actions.action.Action` succeeded or failed.

        Notes
        -----
        This method is called by the :class:`matrx.grid_world.GridWorld` iff:

        * An :class:`matrx.agents.agent_brain.AgentBrain` of an agent decided
          upon this action.
        * It is the turn of the agent to take an action.
        * The implemented :meth:`matrx.actions.Action.is_possible` method
          returned an :class:`matrx.actions.Action.ActionResult.succeeded`
          denoting success.
        * The `action_duration` has passed.

        Furthermore, it is important to remember that the given
        :class:matrx.grid_world.GridWorld` instance, represents the actual
        world. Any change to this instance are immediately reflected in the
        world.

        See Also
        --------
        matrx.grid_world.GridWorld :
            Contains additional functions to make the implementation of
            :class:`matrx.actions.action.Action` simpler.
        ActionResult :
            Contains a description on how to create an instance of
            :class:`matrx.actions.action.ActionResult`, or how to override it
            for your own custom :class:`matrx.actions.action.Action`.

        """
        return None

    def is_possible(self, grid_world, agent_id, **kwargs):
        """ Checks if the Action is possible.

        This method analyses a :class:`matrx.grid_world.GridWorld` instance
        to check if the Action is possible to perform. Override this method
        when implementing a new :class:`matrx.actions.action.Action`.

        Parameters
        ----------
        grid_world : GridWorld
            The GridWorld instance that should be used to check if this Action
            is possible.
        agent_id : string
            The unique identifier of the agent performing this action.
        kwargs : dictionary
            The set of keyword arguments provided by the Agent that decided
            upon this action. When overriding this method and setting required
            arguments, the GridWorld will take responsibility in checking
            whether the Agent provided these. Hence, it is not required to do
            so here.

        Returns
        -------
        ActionResult
            The expected :class:`matrx.actions.action.ActionResult` when
            performing this :class:`matrx.actions.action.Action`. The
            :attr:`matrx.actions.action.ActionResult.succeeded` attribute is
            used to check if the :class:`matrx.actions.action.Action` is
            indeed possible by the :class:`matrx.grid_world.GridWorld`.

        Notes
        -----
        It is important to understand that this function does not have to
        guarantee the success of an :class:`matrx.actions.action.Action`. It
        might be impossible to check if an
        :class:`matrx.actions.action.Action` would always succeed without
        actually performing the mutation. Hence, this method only needs to
        check for general cases. The
        :meth:`matrx.actions.action.Action.mutate` itself also returns an
        :class:`matrx.actions.action.ActionResult` which denotes the actual
        success or failure of the Action.

        This method is called by the :class:`matrx.grid_world.GridWorld` iff:

        * An :class:`matrx.agents.agent_brain.AgentBrain` of an agent decided
          upon this action.
        * It is the turn of the agent to take an action.
        * The `action_duration` has passed.

        OR

        * An :class:`matrx.agents.agent_brain.AgentBrain` requested if a
          certain :class:`matrx.actions.action.Action` is possible. (Requested
          functionality: See Issue #46)

        Warnings
        --------
        This method should not change anything in this instance. This is
        reserved for the :meth:`matrx.actions.action.Action.mutate` method as
        any change here will be reflected in the world.

        See Also
        --------
        GridWorld :
            Contains additional functions to make the implementation of Actions
            simpler.
        ActionResult :
            Contains a description on how to create an ActionResult.

        """
        return None


class ActionResult:
    """ Represents the generic (expected) result of an action.

    This class functions as a simple wrapper around a boolean representing the
    success (`True`) or failure (`False`) of the (expected) mutation of an
    :class:`matrx.actions.action.Action`, as well as the reason for it.

    The :class:`matrx.actions.action.ActionResult` class contains several
    generic reasons for a succeed or fail as constant class attributes.

    Parameters
    ----------
    result : str
        A string representing the reason for the (expected) success or fail of
        an :class:`matrx.actions.action.Action`.
    succeeded : bool
        A boolean representing the (expected) success or fail of an
        :class:`matrx.actions.action.Action`.

    Notes
    -----
    Both the methods :meth:`matrx.actions.Action.is_possible` and
    :meth:`matrx.actions.Action.mutate` should return an
    :class:`matrx.actions.action.ActionResult`. With the former it represents
    the expected success or failure of the
    :class:`matrx.actions.action.Action`, with the latter the actual success
    or failure.

    The :class:`matrx.actions.action.ActionResult` class contain several
    reasons why an :class:`matrx.actions.action.Action` may succeed or fail
    that are being checked by a :class:`matrx.grid_world.GridWorld` instance.
    These are:

        * IDLE_ACTION: Always given when an AgentBrain decided upon returning
          None (the idle action)
        * AGENT_WAS_REMOVED: Always given when an agent does not exist anymore
          in the GridWorld.
        * AGENT_NOT_CAPABLE: Always given when the Action an AgentBrain decided
          upon is not an Action that agent
          should be capable of.
        * UNKNOWN_ACTION: Always given when the Action name an AgentBrain
          decided upon is not recognized as a class name of an Action or some
          class inheriting from Action. Could be caused by either that class
          not existing or because it is never imported anywhere (and hence is
          not on the path the GridWorld searches in).

    When creating a new :class:`matrx.actions.action.Action` class, you can
    also extend this class with your own. This allows you to set your own
    constants representing reasons for that
    :class:`matrx.actions.action.Action` specific results. This allows an
    :class:`matrx.agents.agent_brain.AgentBrain` to directly match the
    :attr:`matrx.actions.ActionResult.result` with a constant and potentially
    adjust its behavior.

    See Also
    --------
    Action : The super-class whose children return an ActionResult

    """

    """ Result string for when the action is None. """
    IDLE_ACTION = "The action is None, hence the agent will idle which " \
                  "always succeeds. "

    """ Result string for when the agent was removed before performing 
    its action. """
    AGENT_WAS_REMOVED = "Agent {AGENT_ID} was removed during this tick, " \
                        "cannot perform anymore actions. "

    """ Result when the action succeeded or is expected to succeed."""
    ACTION_SUCCEEDED = "The action succeeded."

    """ Result when the action is not expected to succeed. """
    ACTION_NOT_POSSIBLE = "The `is_possible(...)` method returned False. " \
                          "Signalling that the action was not possible. "

    """ Result when the agent is not capable of performing that action. """
    AGENT_NOT_CAPABLE = "The action could not be performed, as the agent is " \
                        "not capable of performing this action. "

    """ Result when the action name does not represent an ``Action`` class"""
    UNKNOWN_ACTION = "The action is not known to the environment."

    """ *Deprecated: Use IDLE_ACTION instead.* """
    NO_ACTION_GIVEN = "There was no action given to perform, automatic " \
                      "succeed. "

    def __init__(self, result, succeeded):
        self.result = result
        self.succeeded = succeeded



import copy
from collections import Iterable, MutableMapping

from matrx import utils
from matrx.objects import Door, AreaTile, Wall
import numpy as np


class State(MutableMapping):

    def __init__(self, own_id, memorize_for_ticks=None):
        if memorize_for_ticks is None:
            self.__decay_val = 0
        else:
            self.__decay_val = 1.0 / memorize_for_ticks

        self.__me = None
        self.__own_id = own_id
        self.__state_dict = {}
        self.__prev_state_dict = {}
        self.__decays = {}

    def state_update(self, state_dict):

        # If decay does not matter, we simply use the given dictionary
        if self.__decay_val <= 0.0:
            # Set the previous and new state
            self.__prev_state_dict = self.__state_dict.copy()
            self.__state_dict = state_dict.copy()

            # Set the "me"
            self.__me = self.get_self()

            # Return self
            return self

        # Else: decay does matter so we need to handle knowledge decay
        prev_state = self.__state_dict.copy()
        state = state_dict.copy()

        # Get the ids of all newly perceived objects
        new_ids = set(state.keys()) - set(prev_state.keys())

        # Get the ids of all objects that were perceived before as well
        persistent_ids = set(state.keys()) - new_ids

        # Get the ids of all objects that are not perceived any more
        gone_ids = set(prev_state.keys()) - set(state.keys())

        # Decay the decays of all objects that are not perceived any longer
        for obj_id in gone_ids:
            self.__decays[obj_id] = max((self.__decays[obj_id] - self.__decay_val), 0)

        # Reset the decays of all objects that are still perceived
        for obj_id in persistent_ids:
            self.__decays[obj_id] = 1.0

        # Add all new decays of the newly perceived objects
        for obj_id in new_ids:
            self.__decays[obj_id] = 1.0

        # Check for non-zero decays and flag them for keeping (this now also
        # includes all new_ids as we just added them).
        to_keep_ids = []
        for obj_id, decay in self.__decays.items():
            if decay > 0:
                to_keep_ids.append(obj_id)
            # remove all zero decay objects, this reduces
            # the self.__decays of growing with zero decays
            else:
                self.__decays.pop(obj_id)
        to_keep_ids = set(to_keep_ids)

        # Create new state
        new_state = {}
        for obj_id in to_keep_ids:
            # If the object id is in the received state, pick that one. This
            # makes sure that any objects that were also in the previous state
            # get updated with newly perceived properties
            if obj_id in state.keys():
                new_state[obj_id] = state[obj_id]
            # If the object id is not in the received state, it may still be
            # in the previous state (e.g. because its decay was non-zero).
            # Though this only happens when decay is set.
            elif obj_id in prev_state.keys():
                new_state[obj_id] = prev_state[obj_id]

        # Set the new state
        self.__prev_state_dict = self.__state_dict
        self.__state_dict = new_state

        # Set the "me"
        self.__me = self.get_self()

        # Return self
        return self

    ###############################################
    # Methods that allow State to be used as dict #
    ###############################################
    def __getitem__(self, key):
        """ Returns all state objects that comply with the given key.

        This method overrides a dict's __getitem__, used in bracket notation, e.g. some_dict[key]. It allows us to
        create a dict that can obtain its items based on more than a single key. It supports the following as keys:

        - A regular key in `state.keys()`. If the key is found, returns that single object.
        - An iterable of keys in `state.keys()`. If *all* keys are found, returns all objects.
        - A string representing a property name. Returns a list of all objects with that property.
        - An iterable of property names. Returns a list of objects that contain all given properties.
        - A dictionary of type {property_name: property_value, ...}. Returns all objects that have all those specified
        property names with their respective property value.
        - A dictionary of type {property_name: [property_value, ...] ...}. Returns all objects that have all those
        specified property names and one of their respective given property value options. This can be mixed with
        property names and values as list or single value (e.g. {name_1: [value_1a, value_1b], name_2: value_2}.

        This allows for a highly versatile search in the perceived state. However, its complexity might be daunting and
        State offers numerous helper methods that make your life simpler. For instance, state.get_with_property(props)
        wraps this method and calls `self[props]

        Parameters
        ----------
        key : str, list, dict
            The key for which we search the state. When a string, it first assumes it is an object ID and if it can't
            find any, assumes it is a property name and finds all objects that have it. If it is a list, it first
            assumes a list of object IDs if not all are object IDs it assumes a list of property names. If a dict, it
            assumes it is of shape {property_name: property_value, ...} or {property_name: [allowable_value, ...], ...}.

        Returns
        -------
        dict, list
            Returns a dict representing the object if only one object was found. If more objects are found, returns a
            list of them. Returns None when no object was found.

        Raises
        ------
        KeyError
            When nothing can be found.

        Examples
        --------
        The examples below assume a world containing several rooms named room_0, room_1 and room_2 containing tiles,
        walls and doors. There are also some other agents and SquareBlock objects.

        Find the agent with the ID 'agent_0`.
        >>> state["agent_0"]

        Find the blocks with the IDs of 'block_321' and 'block_543'
        >>> state[['block_321', 'block_543']]

        Find all objects with the property 'is_open' (e.g. doors)
        >>> state['is_open']

        Find all objects with the property 'is_open' and value 'True' (all open doors)
        >>> state[{'is_open': True}]

        Find all objects with the property 'room_name' and value 'room_0' but also with the property 'class_inheritance'
         and value 'Wall' (e.g. all walls of room_0)
        >>> state['room_name': 'room_0', 'class_inheritance': 'Wall']

        Find all open doors of room_0.
        >>> state[{'room_name': 'room_0', 'class_inheritance': 'Door', 'is_open': 'True'}]

        Find all objects that are either a Wall or Door.
        >>> state['class_inheritance': ['Wall', 'Door']]

        It works for any property, including custom properties: Given a custom property 'foo' and a possible value 'b',
        'a' or 'r', find all objects with value 'b' and 'a'.
        >>> state['foo': ['b', 'a']]

        It works for any possible value (except for dict, see Notes).
        >>> state['number': [1, 2], 'some_list': [['a', 1], ['b', 2]]]

        Notes
        -----
        In case a passed property value is an iterable, it searches whether that iterable is *part* of an objects value
        for that property.

        E.g. state[{'some_list': ['a', 1]}] will also return an object with obj['some_list'] = ['a', 'b', 1, 2]}

        Warnings
        --------
        You cannot find any objects with a property name whose values are a dict. This means you cannot (as of now) use
        this method to find all objects with a certain colour using something as:

        `state[{'visualization': {'colour': '#000000'}}]`

        Instead of returning all objects that are black, it returns all objects with a 'visualization' property that
        contains the key 'colour' (which is basically every object). All objects that do not have the 'visualization'
        property or do have it but not with the key 'colour' in it, are ignored.

        The 'visualization' property is one such example. But MATRX allows to add custom properties whose values are
        dicts. Currently, there is no method to quickly find these using State. Instead you should rely on finding it
        yourself.

        If you want to locate objects with certain visualization properties, we offer several helper methods to do so.
        These are; state.get_with_colour(...), state.get_with_size(...), state.get_with_shape(...),
        state.get_with_depth(...), and state.get_with_opacity(...).
        """
        found_objects = self.__find_object(props=key, combined=True)
        if found_objects is not None and len(found_objects) == 1:  # just a single object
            return found_objects[0]
        return found_objects

    def __setitem__(self, key, value):
        raise ValueError("You cannot set items to the state, use state.state_update(...) instead.")

    def __delitem__(self, key):
        del self.__state_dict[key]

    def __iter__(self):
        return iter(self.__state_dict)

    def __len__(self):
        return len(self.__state_dict)

    def update(self, *args, **kwargs):
        raise ValueError("You cannot update the state, use state.state_update(...) instead.")

    def copy(self):
        return copy.deepcopy(self)

    def pop(self, obj_id):
        return self.__state_dict.pop(obj_id)

    def remove(self, obj_id):
        self.__state_dict.pop(obj_id)

    def as_dict(self):
        return self.__state_dict

    ###############################################
    #     Some helpful getters for the state      #
    ###############################################
    def get_with_property(self, props, combined=True):
        found = self.__find_object(props, combined)
        return found

    def get_world_info(self):
        return self.__state_dict['World']

    def remove_with_property(self, props, combined=True):
        found = self.__find_object(props, combined)
        if found is None:
            return
        for obj in found:
            self.remove(obj['obj_id'])

    def get_of_type(self, obj_type):
        return self.get_with_property({"class_inheritance": obj_type}, combined=False)

    def get_room(self, room_name):
        return self.get_with_property({"room_name": room_name}, combined=False)

    def get_all_room_names(self):
        return list({obj['room_name'] for obj in self.get_with_property("room_name", combined=False)})

    def get_room_objects(self, room_name):
        # Locate method to identify room content
        def is_content(obj):
            if 'class_inheritance' in obj.keys():
                chain = obj['class_inheritance']
                if not (Wall.__name__ in chain or Door.__name__ in chain or AreaTile in chain) \
                        and obj['room_name'] == room_name:
                    return obj
            else:  # the object is a Wall, Door or AreaTile
                return None

        # Get all room objects
        room_objs = self.get_room(room_name)

        if room_objs is None:  # No room with room_name was found
            return None

        # Filter out all area's, walls and doors
        content = map(is_content, room_objs)
        content = [c for c in content if c is not None]

        return content

    def get_room_doors(self, room_name):
        # Locate method to identify doors of the right room
        def is_content(obj):
            if 'class_inheritance' in obj.keys():
                chain = obj['class_inheritance']
                if Door.__name__ in chain and obj['room_name'] == room_name:
                    return obj
            else:  # the object is not a Door
                return None

        room_objs = self.get_room(room_name)
        if room_objs is None:  # No room was found with the given room name
            return None

        # Filter out all doors
        doors = map(is_content, room_objs)
        doors = [c for c in doors if c is not None]

        return doors

    def get_agents(self):
        agents = self.__find_object(props={'class_inheritance': 'AgentBrain'}, combined=False)
        return agents

    def get_agents_with_property(self, props):
        if isinstance(props, str):
            props = {'class_inheritance': 'AgentBrain', props: (None, )}
        else:
            props = {'class_inheritance': 'AgentBrain', **props}
        agents = self.__find_object(props=props, combined=True)
        return agents

    def get_team_members(self, team_name=None):
        if team_name is None:
            if self.__me is not None:
                team_name = self.__me['team']
            else:
                team_name = self.get_self()['team']
        team_members = self.get_agents_with_property({'team': team_name})
        return team_members

    def get_closest_objects(self):
        objs = self.__find_object(props='location', combined=True)
        if objs is None:
            return None

        objs = np.array(objs)
        # Remove itself, since the agent self is always closest...
        other_objects = np.array([o for o in objs if o['obj_id'] != self.__own_id])
        closest_objects = self.__get_closest(other_objects)
        return closest_objects

    def get_closest_with_property(self, props):
        objs = self.__find_object(props=props, combined=True)
        if objs is None:
            return None

        objs = np.array(objs)
        # Remove itself, since the agent self is always closest...
        other_objects = np.array([o for o in objs if o['obj_id'] != self.__own_id])
        closest_objects = self.__get_closest(other_objects)
        return closest_objects

    def get_closest_room_door(self, room_name=None):
        if room_name is None:
            objs = self.get_with_property({"class_inheritance": "Door", "room_name": room_name}, combined=True)
        else:
            objs = self.get_room_doors(room_name)

        if objs is None:
            return None

        objs = np.array(objs)
        closest_objects = self.__get_closest(objs)
        return closest_objects

    def get_closest_agents(self):
        agents = self.get_agents()
        if agents is None:
            return None

        # Remove itself, since the agent self is always closest...
        other_agents = np.array([a for a in agents if a['obj_id'] != self.__own_id])
        closest_agents = self.__get_closest(other_agents)
        return closest_agents

    def get_self(self):
        me = self.__find_object(props={'obj_id': self.__own_id}, combined=True)
        return me

    ###############################################
    # Some higher level abstractions of the state #
    ###############################################
    def get_traverse_map(self):
        width, length = self.get_world_info()['grid_shape']
        coords = [(x, y) for x in range(width) for y in range(length)]
        traverse_map = {c: True for c in coords}
        intrav_objs = self.get_with_property(props={'is_traversable': False, 'location': (None, )}, combined=False)
        for o in intrav_objs:
            traverse_map[(o['location'][0], o['location'][1])] = False

        return traverse_map

    def get_distance_map(self):
        width, length = self.get_world_info()['grid_shape']
        coords = [(x, y) for x in range(width) for y in range(length)]
        if self.__me is not None:
            loc = self.__me['location']
        else:
            loc = self.get_self()['location']

        def distance(coord):
            return utils.get_distance(loc, coord)

        dist_map = {c: distance(c) for c in coords}

        return dist_map

    def apply_occlusion(self):
        raise NotImplemented("Field of view occlusion is not yet implemented.")

    ##################################################
    # The basic functions that make up most of state #
    ##################################################
    def __get_closest(self, objs):
        if len(objs) == 0:
            return None

        if self.__me is not None:
            my_loc = self.__me['location']
        else:
            my_loc = self.get_self()['location']

        def distance(x):
            loc = x['location']
            return utils.get_distance(my_loc, loc)

        dists = np.array(list(map(distance, objs)))
        closest_idx = np.where(dists == dists.min())[0]
        closest_objects = objs[closest_idx]
        closest_objects = list(closest_objects)

        return closest_objects

    def __find_object(self, props, combined):
        # Make sure that props is a dict, with as keys the property names and as values a tuple of allowable property
        # values (which can be (None,) if no value is specified).
        if isinstance(props, dict):
            # if props is a dict, we check if its values are tuples and cast them to tuples when there are any iterable
            # and wrap them in a tuple if it is a single value.
            props = {p: v if isinstance(v, tuple) else tuple(v) if State.__is_iterable(v) else tuple([v])
                     for p, v in props.items()}
        elif isinstance(props, str):  # props is a single string
            # It could be that props is in fact an "obj_id", the only value allowed to be passed and return something so
            # we check if it is in our dict and return it.
            if props in self.__state_dict.keys():
                found = [self.__state_dict[props]]
                return found
            # props is a single property, so make a appropriate dict out of it with no value
            props = {props: (None, )}
        elif State.__is_iterable(props):
            # props is a list, and it may be a list of object ids so first check if this is the case
            found = [self.__state_dict[obj_id] for obj_id in props if obj_id in self.__state_dict.keys()]
            if len(found) != len(props):  # if one given prop_name was not an actual obj_id, treat all as property names
                # props is a list, assuming a list of property names.
                props = {p: (None,) for p in props}
            else:  # all property names were in fact keys, so return what we found
                return found

        # For each prop_name, prop_value combination, find the relevant objects. If there are more than one allowable
        # property value, search for those as well.
        found = [[self.__find(name, val) for val in vals] for name, vals in props.items()]

        # Check if we searched for more than one property
        if len(props) > 1:
            # If we just want all objects with EITHER property (potentially with the set value), we just flatten the
            # list (prop names) of lists (prop values).
            if not combined:
                # we know we have multiple property names, so we flatten in that dimension
                found = [value_found for sub_found in found for value_found in sub_found]
                # know we also flatten in the property value dimension, resulting in one big list of all objects that
                # have the specified property name and property values
                found = [obj for value_found in found for obj in value_found]

            # If we want all objects that have ALL the properties (potentially also with their respective value), we
            # select those objects that are in all sub-lists
            else:
                # We transform all lists of objects that adhere to each property name (and possible value) to a set
                sets = [set(obj['obj_id'] for obj_list in sub_found for obj in obj_list) for sub_found in found]

                # Next we loop over each set, and consequtively take the intersection of each set. E.g. we start with
                # the first set as the intersection, and set it to the intersection of the previous intersect and next
                # set,
                intersect = set(sets[0])
                for s in sets[1:]:
                    intersect = intersect.intersection(s)

                # Now we have all the object ids, but we want the objects. So if the intersection is not empty, we
                # transform all found objects to a single dict with their ids as keys. Next we loop through every id and
                # retrieve the object dict that belongs to it.
                if len(intersect) > 0:
                    full_dict = {obj['obj_id']: obj for sub_found in found for obj_list in sub_found for obj in
                                 obj_list}
                    found = [full_dict[obj_id] for obj_id in intersect]
                else:
                    found = []
        else:  # just one property name given
            found = found[0]  # take that property name list
            if len(list(props.values())[0]) > 1:  # we have more than one property values given
                # flatten the list
                found = [obj for f in found for obj in f]
            elif len(found[0]) == 0:  # if just one property value was given (or omitted) but nothing was found
                found = None
            elif len(found) > 0 and len(found[0]) > 0:
                found = found[0]

        # If nothing was found, we set it to None for easy identification and break any iterable over it.
        if not found:
            found = None

        return found

    def __find(self, prop_name, prop_value=None):
        # A local function that identifies whether a given obj_id-obj pair has the requested property name and, if
        # given, the right property value. Is used in the map method to find all object that adhere to it.
        def locate(id_obj_pair):
            obj = id_obj_pair[1]

            # The object is found when it has the property and the value is none, or when it has the property and the
            # requested value is the value of that property OR is in that value of that property (e.g. as substring or
            # list item).
            if prop_name in obj:
                # value given and equals that in the object
                if prop_value is None or prop_value == obj[prop_name]:
                    return obj
                # value given and is in that object's property (e.g. as substring of item in list)
                elif prop_value is not None and isinstance(obj[prop_name], Iterable) and prop_value in obj[prop_name]:
                    return obj
            return None

        # try to locate it in the regular dict as an object id, this allows us to use this method also to quickly locate
        # objects based on their object ID.
        try:
            return self.__state_dict[prop_value]
        # if the prop value was not a key in the dict (or simply None), find all objects with that property and value if
        # given. This uses Python's map function to bring our search to C.
        except KeyError:
            located = map(locate, self.__state_dict.items())
            return [l for l in located if l is not None]  # only return the found objects

    @staticmethod
    def __is_iterable(arg):
        # Checks if the arg functions as an iterable (e.g. is a list, tuple, set, dict, etc.). The isinstance method
        # would be less specific or very large to include all desirable types. Since, isinstance(arg, Iterable) would
        # also pass for any strings, but isinstance(arg, (list, tuple, dict, set, array, ...)) grows quite large.
        return not hasattr(arg, "strip") and (hasattr(arg, "__getitem__") or hasattr(arg, "__iter__"))



from matrx.logger.logger import GridWorldLogger


class LogDuration(GridWorldLogger):

    def __init__(self, save_path="", file_name_prefix="", file_extension=".csv", delimeter=";"):
        super().__init__(save_path=save_path, file_name=file_name_prefix, file_extension=file_extension,
                         delimiter=delimeter, log_strategy=self.LOG_ON_LAST_TICK)

    def log(self, grid_world, agent_data):
        log_statement = {
            "tick": grid_world.current_nr_ticks
        }

        return log_statement



from matrx.logger.logger import GridWorldLogger


class LogIdleAgents(GridWorldLogger):

    def __init__(self, log_strategy=1, save_path="", file_name_prefix="", file_extension=".csv", delimeter=";"):
        super().__init__(log_strategy=log_strategy, save_path=save_path, file_name=file_name_prefix,
                         file_extension=file_extension, delimiter=delimeter)

    def log(self, grid_world, agent_data):
        log_statement = {}
        for agent_id, agent_obj in grid_world.registered_agents.items():
            idle = agent_obj.properties['current_action'] is None
            log_statement[agent_id] = int(idle)

        for agent_id, agent_obj in grid_world.registered_agents.items():
            if log_statement[agent_id] != 0:
                return log_statement

        return None



import collections

import numpy as np

from matrx.actions.action import Action, ActionResult
from matrx.utils import get_distance
from matrx.objects.agent_body import AgentBody
from matrx.objects.standard_objects import AreaTile


class RemoveObject(Action):
    """ Removes an object from the world.

    An action that permanently removes an
    :class:`matrx.objects.env_object.EnvObject` from the world, which can be
    any object except for the agent performing the action.

    Parameters
    ----------
    duration_in_ticks : int
        Optional. Default: ``1``. Should be zero or larger.

        The default duration of this action in ticks during which the
        :class:`matrx.grid_world.GridWorld` blocks the agent performing other
        actions. By default this is 1, meaning that this action will take both
        the tick in which it was decided upon and the subsequent tick.

    """

    def __init__(self, duration_in_ticks=0):
        super().__init__(duration_in_ticks)

    def mutate(self, grid_world, agent_id, **kwargs):
        """ Removes the specified object.

        Removes a specific :class:`matrx.objects.env_object.EnvObject` from
        the world. Can be any object except for the agent performing the
        action.

        Parameters
        ----------
        grid_world : GridWorld
            The ``matrx.grid_world.GridWorld`` instance in which the object is
            sought according to the `object_id` parameter.
        agent_id : str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id : str
            Optional. Default: ``None``

            The string representing the unique identifier of the
            :class:`matrx.objects.env_object.EnvObject` that should be
            removed.
        remove_range : int
            Optional. Default: ``1``

            The range in which the to be removed
            :class:`matrx.objects.env_object.EnvObject` should be in.

        Returns
        -------
        RemoveObjectResult
            Depicts the action's success or failure and reason for that result.

            See :class:`matrx.actions.object_actions.RemoveObjectResult` for
            the results it can contain.

        """
        assert 'object_id' in kwargs.keys()  # assert if object_id is given.
        object_id = kwargs['object_id']  # assign
        remove_range = 1  # default remove range
        if 'remove_range' in kwargs.keys():  # if remove range is present
            assert isinstance(kwargs['remove_range'], int)  # should be of integer
            assert kwargs['remove_range'] >= 0  # should be equal or larger than 0
            remove_range = kwargs['remove_range']  # assign

        # get the current agent (exists, otherwise the is_possible failed)
        agent_avatar = grid_world.registered_agents[agent_id]
        agent_loc = agent_avatar.location  # current location

        # Get all objects in the remove_range
        objects_in_range = grid_world.get_objects_in_range(agent_loc, object_type="*", sense_range=remove_range)

        # You can't remove yourself
        objects_in_range.pop(agent_id)

        for obj in objects_in_range:  # loop through all objects in range
            if obj == object_id:  # if object is in that list
                success = grid_world.remove_from_grid(object_id)  # remove it, success is whether GridWorld succeeded
                if success:  # if we succeeded in removal return the appropriate ActionResult
                    return RemoveObjectResult(RemoveObjectResult.OBJECT_REMOVED.replace('object_id'.upper(),
                                                                                        str(object_id)), True)
                else:  # else we return a failure due to the GridWorld removal failed
                    return RemoveObjectResult(RemoveObjectResult.REMOVAL_FAILED.replace('object_id'.upper(),
                                                                                        str(object_id)), False)

        # If the object was not in range, or no objects were in range we return that the object id was not in range
        return RemoveObjectResult(RemoveObjectResult.OBJECT_ID_NOT_WITHIN_RANGE
                                  .replace('remove_range'.upper(), str(remove_range))
                                  .replace('object_id'.upper(), str(object_id)), False)

    def is_possible(self, grid_world, agent_id, **kwargs):
        """ Checks if an object can be removed.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            object is sought according to the `object_id` parameter.
        agent_id: str
            The string representing the unique identified that represents the
            agent performing this action.
        object_id: str
            Optional. Default: ``None``

            The string representing the unique identifier of the
            :class:`matrx.objects.env_object.EnvObject` that should be
            removed.
        remove_range : int
            Optional. Default: ``1``

            The range in which the to be removed
            :class:`matrx.objects.env_object.EnvObject` should be in.

        Returns
        -------
        RemoveObjectResult
            The :class:`matrx.actions.action.ActionResult` depicting the
            action's expected success or failure and reason for that result.

            See :class:`matrx.actions.object_actions.RemoveObjectResult` for
            the results it can contain.

        """
        agent_avatar = grid_world.get_env_object(agent_id, obj_type=AgentBody)  # get ourselves
        assert agent_avatar is not None  # check if we actually exist
        agent_loc = agent_avatar.location  # get our location

        remove_range = np.inf  # we do not know the intended range, so assume infinite
        # get all objects within infinite range
        objects_in_range = grid_world.get_objects_in_range(agent_loc, object_type="*", sense_range=remove_range)

        # You can't remove yourself
        objects_in_range.pop(agent_avatar.obj_id)

        if len(objects_in_range) == 0:  # if there are no objects in infinite range besides ourselves, we return fail
            return RemoveObjectResult(RemoveObjectResult.NO_OBJECTS_IN_RANGE.replace('remove_range'.upper(),
                                                                                     str(remove_range)), False)

        # otherwise some instance of RemoveObject is possible, although we do not know yet IF the intended removal is
        # possible.
        return RemoveObjectResult(RemoveObjectResult.ACTION_SUCCEEDED, True)


class RemoveObjectResult(ActionResult):
    """ActionResult for a RemoveObjectAction

    The results uniquely for RemoveObjectAction are (as class constants):

    * OBJECT_REMOVED: If the object was successfully removed.
    * REMOVAL_FAILED: If the object could not be removed by the
      :class:`matrx.grid_world.GridWorld`.
    * OBJECT_ID_NOT_WITHIN_RANGE: If the object is not within specified range.
    * NO_OBJECTS_IN_RANGE: If no objects are within range.

    Parameters
    ----------
    result: str
        A string representing the reason for the (expected) success or fail of
        a :class:`matrx.actions.object_actions.RemoveObjectAction`.
    succeeded: bool
        A boolean representing the (expected) success or fail of a
        :class:`matrx.actions.object_actions.RemoveObjectAction`.

    See Also
    --------
    RemoveObjectAction

    """

    """ Result when the specified object is successfully removed. """
    OBJECT_REMOVED = "The object with id `OBJECT_ID` is removed."

    """ Result when no objects were within the specified range. """
    NO_OBJECTS_IN_RANGE = "No objects were in `REMOVE_RANGE`."

    """ Result when the specified object is not within the specified range. """
    OBJECT_ID_NOT_WITHIN_RANGE = "The object with id `OBJECT_ID` is not within the range of `REMOVE_RANGE`."

    """ Result when the world could not remove the object for some reason. """
    REMOVAL_FAILED = "The object with id `OBJECT_ID` failed to be removed by the environment for some reason."

    def __init__(self, result, succeeded):
        super().__init__(result, succeeded)


class GrabObject(Action):
    """ Grab and hold objects.

    The action that can pick up / grab and hold an
    :class:`matrx.objects.env_object.EnvObject`. Cannot be performed on agents
    (including the agent performing the action). After grabbing / picking up,
    the object is automatically added to the agent's inventory and removed from
    the :class:`matrx.grid_world.GridWorld`.

    Parameters
    ----------
    duration_in_ticks : int
        Optional. Default: ``1``

        The default duration of the action in ticks during which the
        :class:`matrx.grid_world.GridWorld` blocks the agent performing other
        actions. By default this is 1, meaning that the action will take both
        the tick in which it was decided upon and the subsequent tick.

    Notes
    -----
    The actual carrying mechanism of objects is implemented in the
    :class:`matrx.actions.move_actions.Move` actions: whenever an agent moves
    who holds objects, those objects it is holding are also moved with it.

    """

    def __init__(self, duration_in_ticks=0):
        super().__init__(duration_in_ticks)

    def is_possible(self, grid_world, agent_id, **kwargs):
        """ Checks if the object can be grabbed.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            object is sought according to the `object_id` parameter.
        agent_id: str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id: str
            Optional. Default: ``None``

            The string representing the unique identifier of the
            :class:`matrx.objects.env_object.EnvObject` that should be
            grabbed. When not given, a random object within range is selected.
        grab_range : int
            Optional. Default: ``np.inf``

            The range in which the to be grabbed
            :class:`matrx.objects.env_object.EnvObject` should be in.
        max_objects : int
            Optional. Default: ``np.inf``

            The maximum of objects the agent can carry.

        Returns
        -------
        GrabObjectResult
            Depicts the action's expected success or failure and reason for
            that result.

            See :class:`matrx.actions.object_actions.GrabObjectResult` for
            the results it can contain.

        """
        # Set default values check
        object_id = None if 'object_id' not in kwargs else kwargs['object_id']
        grab_range = np.inf if 'grab_range' not in kwargs else kwargs['grab_range']
        max_objects = np.inf if 'max_objects' not in kwargs else kwargs['max_objects']

        return _is_possible_grab(grid_world, agent_id=agent_id, object_id=object_id, grab_range=grab_range,
                                 max_objects=max_objects)

    def mutate(self, grid_world, agent_id, **kwargs):
        """ Grabs an object.

        Alters the properties of the agent doing the grabbing, and the object
        being grabbed (and carried), such that the agent's inventory contains
        the entire object and the object being carried properties contains the
        agent's id.

        The grabbed object is removed from the world, and will only exist
        inside of the agent's inventory.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            object is sought according to the `object_id` parameter.
        agent_id: str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id: str
            Optional. Default: ``None``

            The string representing the unique identifier of the
            :class:`matrx.objects.env_object.EnvObject` that should be
            grabbed. When not given, a random object within range is selected.
        grab_range : int
            Optional. Default: ``np.inf``

            The range in which the to be grabbed
            :class:`matrx.objects.env_object.EnvObject` should be in.
        max_objects : int
            Optional. Default: ``np.inf``

            The maximum of objects the agent can carry.

        Returns
        -------
        GrabObjectResult
            Depicts the action's expected success or failure and reason for
            that result.

            See :class:`matrx.actions.object_actions.GrabObjectResult` for
            the results it can contain.

        Notes
        -----
        A grabbed object resides inside the inventory of an agent, not directly
        in the world any longer. Hence, if the agent is removed, so is its
        inventory and all objects herein.

        """

        # Additional check
        assert 'object_id' in kwargs.keys()
        assert 'grab_range' in kwargs.keys()
        assert 'max_objects' in kwargs.keys()

        # if possible:
        object_id = kwargs['object_id']  # assign

        # Loading properties
        reg_ag = grid_world.registered_agents[agent_id]  # Registered Agent
        env_obj = grid_world.environment_objects[object_id]  # Environment object

        # Updating properties
        env_obj.carried_by.append(agent_id)
        reg_ag.is_carrying.append(env_obj)  # we add the entire object!

        # Remove it from the grid world (it is now stored in the is_carrying list of the AgentAvatar
        succeeded = grid_world.remove_from_grid(object_id=env_obj.obj_id, remove_from_carrier=False)
        if not succeeded:
            return GrabObjectResult(GrabObjectResult.FAILED_TO_REMOVE_OBJECT_FROM_WORLD.replace("{OBJECT_ID}",
                                                                                                env_obj.obj_id), False)

        # Updating Location (done after removing from grid, or the grid will search the object on the wrong location)
        env_obj.location = reg_ag.location

        return GrabObjectResult(GrabObjectResult.RESULT_SUCCESS, True)


class GrabObjectResult(ActionResult):
    """ActionResult for a GrabObjectAction

    The results uniquely for GrabObjectAction are (as class constants):

    * RESULT_SUCCESS: When the object can be successfully grabbed.
    * RESULT_NO_OBJECT: When `object_id` is not given.
    * RESULT_CARRIES_OBJECT: When the agent already carries the maximum nr.
      objects.
    * NOT_IN_RANGE: When `object_id` not within range.
    * RESULT_AGENT: If the `object_id` is that of an agent.
    * RESULT_OBJECT_CARRIED: When the object is already carried by another
      agent.
    * RESULT_OBJECT_UNMOVABLE: When the object is not movable.
    * RESULT_UNKNOWN_OBJECT_TYPE: When the `object_id` does not exists in the
      :class:`matrx.grid_world.GridWorld`.
    * FAILED_TO_REMOVE_OBJECT_FROM_WORLD: When the grabbed object cannot be
      removed from the :class:`matrx.grid_world.GridWorld`.

    Parameters
    ----------
    result: str
        A string representing the reason for a
        :class:`matrx.actions.object_actions.GrabObjectAction` (expected)
        success or fail.
    succeeded: bool
        A boolean representing the (expected) success or fail of a
        :class:`matrx.actions.object_actions.GrabObjectAction`.

    See Also
    --------
    GrabObjectAction

    """

    """ Result when the object can be successfully grabbed. """
    RESULT_SUCCESS = 'Grab action success'

    """ Result when the grabbed object cannot be removed from the 
    :class:`matrx.grid_world.GridWorld`. """
    FAILED_TO_REMOVE_OBJECT_FROM_WORLD = 'Grab action failed; could not remove object with id {OBJECT_ID} from grid.'

    """ Result when the specified object is not within range. """
    NOT_IN_RANGE = 'Object not in range'

    """ Result when the specified object is an agent. """
    RESULT_AGENT = 'This is an agent, cannot be picked up'

    """ Result when no object was specified. """
    RESULT_NO_OBJECT = 'No Object specified'

    """ Result when the agent is at its maximum carrying capacity. """
    RESULT_CARRIES_OBJECT = 'Agent already carries the maximum amount of objects'

    """ Result when the specified object is already carried by another agent. 
    """
    RESULT_OBJECT_CARRIED = 'Object is already carried by {AGENT_ID}'

    """ Result when the specified object does not exist in the 
    :class:`matrx.grid_world.GridWorld` """
    RESULT_UNKNOWN_OBJECT_TYPE = 'obj_id is no Agent and no Object, unknown what to do'

    """ Result when the specified object is not movable. """
    RESULT_OBJECT_UNMOVABLE = 'Object is not movable'

    def __init__(self, result, succeeded):
        super().__init__(result, succeeded)


class DropObject(Action):
    """ Drops a carried object.

        The action that can drop an :class:`matrx.objects.env_object.EnvObject`
        that is in an agent's inventory. After dropping, the object is added to the
        :class:`matrx.grid_world.GridWorld`.

        Parameters
        ----------
        duration_in_ticks : int
            Optional. Default: ``1``. Should be zero or larger.

            The default duration of the action in ticks during which the
            :class:`matrx.grid_world.GridWorld` blocks the agent performing other
            actions. By default this is 1, meaning that the action will take both
            the tick in which it was decided upon and the subsequent tick.

        Notes
        -----
        The actual carrying mechanism of objects is implemented in the
        :class:`matrx.actions.move_actions.Move` actions: whenever an agent moves
        who holds objects, those objects it is holding are also moved with it.

        """
    
    def __init__(self, duration_in_ticks=0):
        super().__init__(duration_in_ticks)

    def is_possible(self, grid_world, agent_id, **kwargs):
        """ Checks if the object can be dropped.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            :class:`matrx.objects.env_object.EnvObject` is dropped.
        agent_id: str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id: str
            Optional. Default: ``None``

            The string representing the unique identifier of the
            :class:`matrx.objects.env_object.EnvObject` that should be
            dropped.

            When not given the last object that was grabbed is dropped.
        drop_range : int
            Optional. Default: ``np.inf``

            The range in which the object can be dropped, with the agent's
            location at its center.

        Returns
        -------
        DropObjectResult
            Depicts the action's expected success or failure and reason for
            that result.

            See :class:`matrx.actions.object_actions.DropObjectResult` for
            the results it can contain.

        """
        reg_ag = grid_world.registered_agents[agent_id]

        drop_range = 1 if 'drop_range' not in kwargs else kwargs['drop_range']

        # If no object id is given, the last item is dropped
        if 'object_id' in kwargs:
            obj_id = kwargs['object_id']
        elif len(reg_ag.is_carrying) > 0:
            obj_id = reg_ag.is_carrying[-1]
        else:
            return DropObjectResult(DropObjectResult.RESULT_NO_OBJECT, False)

        return _possible_drop(grid_world, agent_id=agent_id, obj_id=obj_id, drop_range=drop_range)

    def mutate(self, grid_world, agent_id, **kwargs):
        """ Drops the carried object.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            :class:`matrx.objects.env_object.EnvObject` is dropped.
        agent_id: str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id: str
            Optional. Default: ``None``

            The string representing the unique identifier of the
            :class:`matrx.objects.env_object.EnvObject` that should be
            dropped.

            When not given the last object that was grabbed is dropped.
        drop_range : int
            Optional. Default: ``np.inf``

            The range in which the object can be dropped.

        Returns
        -------
        DropObjectResult
            The :class:`matrx.actions.action.ActionResult` depicting the
            action's expected success or failure and reason for that result.

            See :class:`matrx.actions.object_actions.DropObjectResult` for
            the results it can contain.

        Raises
        ------
        Exception
            When the object is said to be dropped inside the agent's location,
            but the agent and object are intraversable. No other intraversable
            objects can be on the same location.

        """
        reg_ag = grid_world.registered_agents[agent_id]

        # fetch range from kwargs
        drop_range = 1 if 'drop_range' not in kwargs else kwargs['drop_range']

        # If no object id is given, the last item is dropped
        if 'object_id' in kwargs:
            env_obj = kwargs['object_id']
        elif len(reg_ag.is_carrying) > 0:
            env_obj = reg_ag.is_carrying[-1]
        else:
            return DropObjectResult(DropObjectResult.RESULT_NO_OBJECT_CARRIED, False)

        # check that it is even possible to drop this object somewhere
        if not env_obj.is_traversable and not reg_ag.is_traversable and drop_range == 0:
            raise Exception(
                f"Intraversable agent {reg_ag.obj_id} can only drop the intraversable object {env_obj.obj_id} at its "
                f"own location (drop_range = 0), but this is impossible. Enlarge the drop_range for the DropAction to "
                f"atleast 1")

        # check if we can drop it at our current location
        curr_loc_drop_poss = _is_drop_poss(grid_world, env_obj, reg_ag.location, agent_id)

        # drop it on the agent location if possible
        if curr_loc_drop_poss:
            return _act_drop(grid_world, agent=reg_ag, env_obj=env_obj, drop_loc=reg_ag.location)

        # if the agent location was the only within range, return a negative action result
        elif not curr_loc_drop_poss and drop_range == 0:
            return DropObjectResult(DropObjectResult.RESULT_OBJECT, False)

        # Try finding other drop locations from close to further away around the agent
        drop_loc = _find_drop_loc(grid_world, reg_ag, env_obj, drop_range, reg_ag.location)

        # If we didn't find a valid drop location within range, return a negative action result
        if not drop_loc:
            return DropObjectResult(DropObjectResult.RESULT_OBJECT, False)

        return _act_drop(grid_world, agent=reg_ag, env_obj=env_obj, drop_loc=drop_loc)


class DropObjectResult(ActionResult):
    """ ActionResult for a DropObjectAction.

    The results uniquely for GrabObjectAction are (as class constants):

    * RESULT_SUCCESS: When the object is successfully dropped.
    * RESULT_NO_OBJECT: When there is no object in the agent's inventory.
    * RESULT_NONE_GIVEN: When the given obj_id is not being carried by the
      agent.
    * RESULT_OBJECT: When the object was intended to drop on the agent's
      location and this was not possible or when no suitable drop location
      could be found.
    * RESULT_UNKNOWN_OBJECT_TYPE: When the object id does not exist (anymore).
    * RESULT_NO_OBJECT_CARRIED: When no objects are carried by the agent.

    Parameters
    ----------
    result: str
        A string representing the reason for the (expected) success or fail of
        an :class:`matrx.actions.object_actions.DropObjectAction`.
    succeeded: bool
        A boolean representing the (expected) success or fail of a
        :class:`matrx.actions.object_actions.DropObjectAction`.

    See Also
    --------
    GrabObjectAction

    """

    """ Result when dropping the object succeeded. """
    RESULT_SUCCESS = 'Drop action success'

    """ Result when there is not object in the agent's inventory. """
    RESULT_NO_OBJECT = 'The item is not carried'

    """ Result when the specified object is not in the agent's inventory. """
    RESULT_NONE_GIVEN = "'None' used as input id"

    """ Result when the specified object should be dropped on an agent. """
    RESULT_AGENT = 'Cannot drop item on an agent'

    """ Result when the specified object should be dropped on an intraversable 
    object."""
    RESULT_OBJECT = 'Cannot drop item on another intraversable object'

    """ Result when the specified object does not exist (anymore). """
    RESULT_UNKNOWN_OBJECT_TYPE = 'Cannot drop item on an unknown object'

    """ Result when the agent is not carrying anything. """
    RESULT_NO_OBJECT_CARRIED = 'Cannot drop object when none carried'

    def __init__(self, result, succeeded, obj_id=None):
        super().__init__(result, succeeded)
        self.obj_id = obj_id


def _is_possible_grab(grid_world, agent_id, object_id, grab_range, max_objects):
    """ Private MATRX method.

    Checks if an :class:`matrx.objects.env_object.EnvObject` can be
    grabbed by an agent.

    Parameters
    ----------
    grid_world : GridWorld
        The :class:`matrx.grid_world.GridWorld` instance in which the
        object is sought according to the `object_id` parameter.
    agent_id: str
        The string representing the unique identified that represents the
         agent performing this action.
    object_id: str
        Optional. Default: ``None``

        The string representing the unique identifier of the
        :class:`matrx.objects.env_object.EnvObject` that should be
        grabbed. When not given, a random object within range is selected.
    grab_range : int
        Optional. Default: ``np.inf``

        The range in which the to be grabbed
        :class:`matrx.objects.env_object.EnvObject` should be in.
    max_objects : int
        Optional. Default: ``np.inf``

        The maximum of objects the agent can carry.

    Returns
    -------
    GrabObjectResult
        Depicts the action's expected success or failure and reason for
        that result.

        Can contain the following results:

        * RESULT_SUCCESS: When the object can be successfully grabbed.
        * RESULT_NO_OBJECT : When `object_id` is not given.
        * RESULT_CARRIES_OBJECT: When the agent already carries the maximum
          nr. objects.
        * NOT_IN_RANGE: When `object_id` not within range.
        * RESULT_AGENT: If the `object_id` is that of an agent.
        * RESULT_OBJECT_CARRIED: When the object is already carried by
          another agent.
        * RESULT_OBJECT_UNMOVABLE: When the object is not movable.
        * RESULT_UNKNOWN_OBJECT_TYPE: When the `object_id` does not exists
          in the :class:`matrx.grid_world.GridWorld`.

    """

    reg_ag = grid_world.registered_agents[agent_id]  # Registered Agent
    loc_agent = reg_ag.location  # Agent location

    if object_id is None:
        return GrabObjectResult(GrabObjectResult.RESULT_NO_OBJECT, False)

    # Already carries an object
    if len(reg_ag.is_carrying) >= max_objects:
        return GrabObjectResult(GrabObjectResult.RESULT_CARRIES_OBJECT, False)

    # Go through all objects at the desired locations
    objects_in_range = grid_world.get_objects_in_range(loc_agent, object_type="*", sense_range=grab_range)
    objects_in_range.pop(agent_id)

    # Set random object in range
    if not object_id:
        # Remove all non objects from the list
        for obj in list(objects_in_range.keys()):
            if obj not in grid_world.environment_objects.keys():
                objects_in_range.pop(obj)

        # Select a random object
        if objects_in_range:
            object_id = grid_world.rnd_gen.choice(list(objects_in_range.keys()))
        else:
            return GrabObjectResult(GrabObjectResult.NOT_IN_RANGE, False)

    # Check if object is in range
    if object_id not in objects_in_range:
        return GrabObjectResult(GrabObjectResult.NOT_IN_RANGE, False)

    # Check if object_id is the id of an agent
    if object_id in grid_world.registered_agents.keys():
        # If it is an agent at that location, grabbing is not possible
        return GrabObjectResult(GrabObjectResult.RESULT_AGENT, False)

    # Check if it is an object
    if object_id in grid_world.environment_objects.keys():
        env_obj = grid_world.environment_objects[object_id]  # Environment object
        # Check if the object is not carried by another agent
        if len(env_obj.carried_by) != 0:
            return GrabObjectResult(GrabObjectResult.RESULT_OBJECT_CARRIED.replace("{AGENT_ID}",
                                                                                   str(env_obj.carried_by)),
                                    False)
        elif not env_obj.properties["is_movable"]:
            return GrabObjectResult(GrabObjectResult.RESULT_OBJECT_UNMOVABLE, False)
        else:
            # Success
            return GrabObjectResult(GrabObjectResult.RESULT_SUCCESS, True)
    else:
        return GrabObjectResult(GrabObjectResult.RESULT_UNKNOWN_OBJECT_TYPE, False)


def _act_drop(grid_world, agent, env_obj, drop_loc):
    """ Private MATRX method.

        Drops the carried object.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            object is dropped.
        agent : AgentBody
            The :class:`matrx.objects.agent_body.AgentBody` of the agent who
            drops the object.
        env_obj : EnvObject
            The :class:`matrx.objects.env_object.EnvObject` to be dropped.
        drop_loc : [x, y]
            The drop location.

        Returns
        -------
        DropObjectResult
            The :class:`matrx.actions.action.ActionResult` depicting the
            action's expected success or failure and reason for that result.

            Returns the following results:

            * RESULT_SUCCESS: When the object is successfully dropped.

        """

    # Updating properties
    agent.is_carrying.remove(env_obj)
    env_obj.carried_by.remove(agent.obj_id)

    # We return the object to the grid location we are standing at
    env_obj.location = drop_loc
    grid_world._register_env_object(env_obj)

    return DropObjectResult(DropObjectResult.RESULT_SUCCESS, True)


def _is_drop_poss(grid_world, env_obj, drop_location, agent_id):
    """ Private MATRX method.

    A breadth first search starting from the agent's location to find the
    closest valid drop location.

    Parameters
    ----------
    grid_world : GridWorld
        The :class:`matrx.grid_world.GridWorld` instance in which the
        object is dropped.
    env_obj : EnvObject
        The :class:`matrx.objects.env_object.EnvObject` to be dropped.
    drop_range : int
        The range in which the object can be dropped.
    start_loc : [x, y]
        The location of the agent from which to start the search.
    agent_id : str
        The agent id of the agent who drops the object.

    Returns
    -------
    boolean
        False if no valid drop location can be found, otherwise the [x,y]
        coordinates of the closest drop location.
    """

    # Count the intraversable objects at the current location if we would drop the
    # object here
    objs_at_loc = grid_world.get_objects_in_range(drop_location, object_type="*", sense_range=0)

    # Remove area objects from the list
    for key in list(objs_at_loc.keys()):
        if AreaTile.__name__ in objs_at_loc[key].class_inheritance:
            objs_at_loc.pop(key)

    # Remove the agent who drops the object from the list (an agent can always drop the
    # traversable object its carrying at its feet, even if the agent is intraversable)
    if agent_id in objs_at_loc.keys():
        objs_at_loc.pop(agent_id)

    in_trav_objs_count = 1 if not env_obj.is_traversable else 0
    in_trav_objs_count += len([obj for obj in objs_at_loc if not objs_at_loc[obj].is_traversable])

    # check if we would have an in_traversable object and other objects in
    # the same location (which is impossible)
    if in_trav_objs_count >= 1 and (len(objs_at_loc) + 1) >= 2:
        return False
    else:
        return True


def _possible_drop(grid_world, agent_id, obj_id, drop_range):
    """ Private MATRX method.

    Checks if an :class:`matrx.objects.env_object.EnvObject` can be
    dropped by an agent.

    Parameters
    ----------
    grid_world : GridWorld
        The :class:`matrx.grid_world.GridWorld` instance in which the
        object is dropped.
    agent_id: str
        The string representing the unique identified that represents the
        agent performing this action.
    obj_id: str
        The string representing the unique identifier of the
        :class:`matrx.objects.env_object.EnvObject` that should be
        dropped.
    drop_range : int
        The range in which the :class:`matrx.objects.env_object.EnvObject`
        should be dropped in.

    Returns
    -------
    DropObjectResult
        The :class:`matrx.actions.action.ActionResult` depicting the
        action's expected success or failure and reason for that result.

        Returns the following results:

        * RESULT_SUCCESS: When the object can be successfully dropped.
        * RESULT_NONE_GIVEN: When the given obj_id is not being carried by
          the agent.
        * RESULT_NO_OBJECT: When no obj_id is given.

    """
    reg_ag = grid_world.registered_agents[agent_id]  # Registered Agent
    loc_agent = reg_ag.location
    loc_obj_ids = grid_world.grid[loc_agent[1], loc_agent[0]]

    # No object given
    if not obj_id:
        return DropObjectResult(DropObjectResult.RESULT_NONE_GIVEN, False)

    # No object with that name
    if not (obj_id in reg_ag.is_carrying):
        return DropObjectResult(DropObjectResult.RESULT_NO_OBJECT, False)

    if len(loc_obj_ids) == 1:
        return DropObjectResult(DropObjectResult.RESULT_SUCCESS, True)

    # TODO: incorporate is_possible check from DropAction.mutate is_possible here

    return DropObjectResult(DropObjectResult.RESULT_SUCCESS, True)


def _find_drop_loc(grid_world, agent, env_obj, drop_range, start_loc):
    """ Private MATRX method.
    A breadth first search starting from the agent's location to find the closest valid drop location.
    Parameters
    ----------
    grid_world : GridWorld
        The GridWorld instance in which the object is dropped.
    agent : AgentBody
        The AgentBody of the agent who drops the object.
    env_obj : EnvObject
        The EnvObject to be dropped.
    drop_range : int
        The range in which the object can be dropped.
    start_loc : [x, y]
        The location of the agent from which to start the search.
    Returns
    -------
    boolean
        False if no valid drop location can be found, otherwise the [x,y] coordinates of the closest drop location.
    """
    queue = collections.deque([[start_loc]])
    seen = {start_loc}

    width = grid_world.shape[0]
    height = grid_world.shape[1]

    while queue:
        path = queue.popleft()
        x, y = path[-1]

        # check if we are still within drop_range
        if get_distance([x, y], start_loc) > drop_range:
            return False

        # check if we can drop at this location
        if _is_drop_poss(grid_world, env_obj, [x, y], agent.obj_id):
            return [x, y]

        # queue unseen neighbouring tiles
        for x2, y2 in ((x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)):
            if 0 <= x2 < width and 0 <= y2 < height and (x2, y2) not in seen:
                queue.append(path + [(x2, y2)])
                seen.add((x2, y2))
    return False



from matrx.agents import AgentBrain
from matrx.agents.agent_utils.navigator import Navigator
from matrx.agents.agent_utils.state_tracker import StateTracker


class PatrollingAgentBrain(AgentBrain):

    def __init__(self, waypoints, move_speed=0):
        super().__init__()
        self.state_tracker = None
        self.navigator = None
        self.waypoints = waypoints
        self.move_speed = move_speed

    def initialize(self):
        # Initialize this agent's state tracker
        self.state_tracker = StateTracker(agent_id=self.agent_id)

        self.navigator = Navigator(agent_id=self.agent_id, action_set=self.action_set,
                                   algorithm=Navigator.A_STAR_ALGORITHM)

        self.navigator.add_waypoints(self.waypoints, is_circular=True)

    def filter_observations(self, state):
        self.state_tracker.update(state)
        return state

    def decide_on_action(self, state):
        from matrx.messages.message import Message

        move_action = self.navigator.get_move_action(self.state_tracker)

        return move_action, {"action_duration": self.move_speed}


import matrx.defaults as defaults
import warnings

class EnvObject:

    def __init__(self, location, name, class_callable, customizable_properties=None,
                 is_traversable=None, is_movable=None,
                 visualize_size=None, visualize_shape=None, visualize_colour=None, visualize_depth=None,
                 visualize_opacity=None, **custom_properties):
        """
        The basic class for all objects in the world. This includes the AgentAvatar. All objects that are added to the
        GridWorld should inherit this class.

        An EnvObject always needs a location and a name. Since we have no idea where you want to put the object or how
        you want to call it.

        In addition agents may alter properties of objects, but only if they are specifically told to be customizable.
        This is done by providing a list of property names in the customizable_properties list.
        Any custom property is allowed to be customizable if in that list, but any keyword argument (e.g.
        is_traversable) may only be influenced by an Action indirectly. For example, the is_traversable property cannot
        be changed in any way, whereas the location property can only be changed indirectly through the PickUpObject
        action.

        A few properties are mandatory for the GridWorld, some Actions and the Visualizer to function. These are
        keyword arguments (e.g. is_traversable). If these are not set, they are obtained from the defaults.py file.

        This class specific allows you to create any object you desire that only differs in the properties it holds. For
        example both a simple block or complex object such as a 'fire' can be modeled with this class. They only differ
        in their mandatory and custom properties (a block is simply a colored square, whereas a 'fire' has potentially
        many more properties). In such a case you simply provide more keyword arguments to the constructor, these
        are automatically added to the self.custom_attributes dictionary. Allowing you to intuitively extend the
        properties of any EnvObject.

        If you want an object that can be altered by an action in a specific way, you can create it as long as it
        inherits from this class. For example the Door object is such an example which has a unique method that 'opens'
        or 'closes' the door. Which is called by the OpenDoor and CloseDoor actions.

        If you want an object that needs to update some of its own properties during the simulation, you can again
        create your own object as long as it inherits from this class. Then you can implement the update_properties
        method. The Battery object is such an example, which simply decreases its energy level property each time step.

        If you have a specific object you need to create a lot and you do not want to keep on setting every property
        every time for the custom_properties, you can make your own class again which must inherit from this class and
        only implement its constructor where these custom properties are set with your default value
        of choosing.

        :param name: String. Mandatory. The name of object, does not need to be unique.
        :param location: List or tuple of length two. Mandatory. The location of the object in the grid world.
        :param customizable_properties: List. Optional, default obtained from defaults.py. The list of attribute names
        that can be customized by other objects (including AgentAvatars and as an extension any Agent).
        :param is_traversable: Boolean. Optional, default obtained from defaults.py. Signals whether other objects can
        be placed on top of this object.
        :param carried_by: List. Optional, default obtained from defaults.py. A list of who is carrying this object.
        :param class_callable: Callable class. Optional, defaults to EnvObject. This is required to make a distinction
        between what kind of object is actually seen or visualized. The last element is always the lowest level class,
        whereas the first element is always EnvObject and everything in between are potential other classes in the
        inheritance chain.
        :param visualize_size: Float. Optional, default obtained from defaults.py. A visualization property used by
        the Visualizer. Denotes the size of the object, its unit is a single grid square in the visualization (e.g. a
        value of 0.5 is half of a square, object is in the center, a value of 2 is twice the square's size centered on
        its location.)
        :param visualize_shape: Int. Optional, default obtained from defaults.py. A visualization property used by the
        Visualizer. Denotes the shape of the object in the visualization. 0=Rectangle, 1=Triangle, 2=Circle
        :param visualize_colour: Hexcode string. Optional, default obtained from defaults.py. A visualization property
        used by the Visualizer. Denotes the colour of the object in visualization.
        :param visualize_depth: Integer. Optional, default obtained from defaults.py. A visualization property that
        is used by the Visualizer to draw objects in layers.
        :param visualize_opacity: Integer. Opacity of the object. From 0.0 to 1.0.
        :param **custom_properties: Optional. Any other keyword arguments. All these are treated as custom attributes.
        For example the property 'heat'=2.4 of an EnvObject representing a fire.
        """

        # Set the object's name.
        self.obj_name = name

        # Obtain a unique ID based on a global object counter, if not already set as an attribute in a super class
        # spaces are not allowed
        if not hasattr(self, "obj_id"):
            self.obj_id = f"{self.obj_name}_{_next_obj_id()}".replace(" ", "_")

            if "#" in self.obj_id:
                warnings.warn("Note: # signs are not allowed as part of an agent or object ID, as it breaks the MATRX frontend. Any hashtags will be removed from the ID..")
                self.obj_id = self.obj_id.replace("#", "")

        # Make customizable_properties mutable if not given.
        if customizable_properties is None:
            self.customizable_properties = []
        else:
            self.customizable_properties = customizable_properties

        # Set the class trace based on the given callable class object. This is required to make a distinction between
        # what kind of object is actually seen or visualized. The last element is always the lowest level class Object,
        # with the second last element being EnvObject. Any elements before that are custom EnvObject class names.
        self.class_inheritance = _get_inheritence_path(class_callable)

        # Load defaults if not given. We do this loading this low-level (
        # instead of for example in the WorldFactory) for users to make it
        # easier to build/extend their own WorldFactory and this way they do
        # not have to deal with this.
        if is_traversable is None:
            is_traversable = defaults.ENVOBJECT_IS_TRAVERSABLE
        if visualize_size is None:
            visualize_size = defaults.ENVOBJECT_VIS_SIZE
        if visualize_shape is None:
            visualize_shape = defaults.ENVOBJECT_VIS_SHAPE
        if visualize_colour is None:
            visualize_colour = defaults.ENVOBJECT_VIS_COLOUR
        if visualize_opacity is None:
            visualize_opacity = defaults.ENVOBJECT_VIS_OPACITY
        if visualize_depth is None:
            visualize_depth = defaults.ENVOBJECT_VIS_DEPTH
        if is_movable is None:
            is_movable = defaults.ENVOBJECT_IS_MOVABLE


        # Set the mandatory properties
        self.visualize_depth = visualize_depth
        self.visualize_colour = visualize_colour
        self.visualize_opacity = visualize_opacity
        self.visualize_shape = visualize_shape
        self.visualize_size = visualize_size
        self.is_traversable = is_traversable
        self.is_movable = is_movable

        # Since carried_by cannot be defined beforehand (it contains the unique id's of objects that carry this object)
        # we set it to an empty list by default.
        self.carried_by = []

        # Go through the custom properties that were given (if any) and set them to the custom_properties dictionary
        self.custom_properties = {}
        for k, v in custom_properties.items():
            self.custom_properties[k] = v

        # location should be set at the end (due to the dependency of its setter on the other properties (e.g. in
        # AgentAvatar)
        self.location = location

    def update(self, grid_world):
        """
        Used to update some properties of this object if needed. For example a 'status' property that changes over time.
        It can also be used to update something in the GridWorld. For example a Fire object that damages other objects
        in its location.

        If you want this functionality, you should create a new object that inherits from this class EnvObject.

        This method is called automatically in the game-loop inside a running GridWorld instance.

        Parameters
        ----------
        grid_world
            The GridWorld instance representing the entire grid world. Can be used to alter itself or others in the
            world in some way.
        """
        pass

    def change_property(self, property_name, property_value):
        """
        Changes the value of an existing (!) property.
        :param property_name: The name of the property.
        :param property_value:  The value of the property.
        :return: The new properties.
        """

        # We check if it is a custom property and if so change it simply in the dictionary
        if property_name in self.customizable_properties:
            self.custom_properties[property_name] = property_value
        else:  # else we need to check if property_name is a mandatory class attribute that is also a property
            if property_name == "is_traversable":
                assert isinstance(property_value, bool)
                self.is_traversable = property_value
            elif property_name == "name":
                assert isinstance(property_value, str)
                self.obj_name = property_value
            elif property_name == "location":
                assert isinstance(property_value, list) or isinstance(property_value, tuple)
                self.location = property_value
            elif property_name == "class_inheritance":
                assert isinstance(property_value, list)
                self.class_inheritance = property_value
            elif property_name == "visualize_size" or property_name == "visualization_size":
                assert isinstance(property_value, int)
                self.visualize_size = property_value
            elif property_name == "visualize_colour" or property_name == "visualization_colour":
                assert isinstance(property_value, str)
                self.visualize_colour = property_value
            elif property_name == "visualize_opacity" or property_name == "visualization_opacity":
                assert isinstance(property_value, float)
                self.visualize_opacity = property_value
            elif property_name == "visualize_shape" or property_name == "visualization_shape":
                assert isinstance(property_value, int)
                self.visualize_shape = property_value
            elif property_name == "visualize_depth" or property_name == "visualization_depth":
                assert isinstance(property_value, int)
                self.visualize_depth = property_value
            elif property_name == "is_movable":
                assert isinstance(property_value, bool)
                self.is_movable = property_value

        return self.properties

    def add_property(self, property_name, property_value):
        """
        Adds a new(!) property with its value to the object.
        """
        if property_name in self.custom_properties:
            raise Exception("Attribute already exists, alter value with change_property instead")
        else:
            # We always add it as a custom property which is also customizable (since we can add it)
            self.custom_properties[property_name] = property_value
            self.customizable_properties.append(property_name)

    @property
    def location(self):
        """
        The location of any object is a pythonic property, this allows us to do various checks on it. One of them is the
        setter that is overridden in AgentAvatar to also transfer all carried objects with it.
        :return: The current location as a tuple; (x, y)
        """
        return tuple(self.__location)

    @location.setter
    def location(self, loc):
        """
        The setter than can be overridden if a location change might also affect some other things inside the
        EnvObject (such as in the case of an AgentAvatar; all the objects its holding change their locations as well).
        :param loc: The new location
        """
        assert isinstance(loc, list) or isinstance(loc, tuple)
        assert len(loc) == 2
        self.__location = loc

    @property
    def properties(self):
        """
        Returns the custom properties of this object, but also any mandatory properties such as location, name,
        is_traversable and all visualization properties (those are in their own dictionary under 'visualization').

        In the case we return the properties of a class that inherits from EnvObject, we check if that class has

        :return: All mandatory and custom properties in a dictionary.
        """

        # Copy the custom properties
        properties = self.custom_properties.copy()

        # Add all mandatory properties. Make sure that these are updated if one are added to the constructor!
        properties['name'] = self.obj_name
        properties['obj_id'] = self.obj_id  # we return id as well, but this should never ever be modified!
        properties['location'] = self.location
        properties['is_movable'] = self.is_movable
        properties['carried_by'] = self.carried_by
        properties['is_traversable'] = self.is_traversable
        properties['class_inheritance'] = self.class_inheritance
        properties['visualization'] = {
            "size": self.visualize_size,
            "shape": self.visualize_shape,
            "colour": self.visualize_colour,
            "depth": self.visualize_depth,
            "opacity": self.visualize_opacity
        }

        return properties

    @properties.setter
    def properties(self, property_dictionary: dict):
        """
        Here to protect the 'properties' variable. It does not do anything and should not do anything!
        """
        pass


object_counter = 0


def _next_obj_id():
    global object_counter
    res = object_counter
    object_counter += 1
    return res


def _get_inheritence_path(callable_class):
    parents = callable_class.mro()
    parents = [str(p.__name__) for p in parents]
    return parents



import heapq
from collections import OrderedDict

import numpy as np
from matrx.agents.agent_utils.state_tracker import StateTracker
from matrx.actions.move_actions import *


class Navigator:
    A_STAR_ALGORITHM = "a_star"

    def __init__(self, agent_id, action_set, algorithm=A_STAR_ALGORITHM, is_circular=False):
        # Set action set
        self.__action_set = action_set

        # Set the path planning algorithm, currently only A* is supported
        self.__algorithm = algorithm

        # Set the move action set the agent is capable of
        self.__move_actions = get_move_actions(action_set)

        # Set the agent ID
        self.__agent_id = agent_id

        # Initialize path plannings algorithm
        self.__path_planning_algo = self.__initialize_path_planner(algorithm=self.__algorithm, action_set=action_set)

        # An ordered dict to store all added waypoints
        self.__waypoints = OrderedDict()

        # Count the current number of waypoints we know of (used as key)
        self.__nr_waypoints = 0

        # Track which waypoint we are currently navigating to (if any)
        self.__current_waypoint_idx = None

        # An ordered dict to store the entire route through all waypoints
        self.__route = OrderedDict()

        # Boolean stating whether the navigator is done
        self.is_done = False

        # Boolean whether this navigator should navigate/patrol in a circular manner through all waypoints
        self.is_circular = is_circular

        # Current traversability map
        self.__occupation_map = None

    def add_waypoint(self, waypoint):
        assert isinstance(waypoint, tuple) or isinstance(waypoint, list)
        wp = Waypoint(loc=waypoint, priority=self.__nr_waypoints)
        self.__nr_waypoints += 1
        self.__waypoints[wp.priority] = wp

    def add_waypoints(self, waypoints, is_circular=False):
        self.is_circular = is_circular

        for waypoint in waypoints:
            self.add_waypoint(waypoint)

    def get_all_waypoints(self):
        return [(k, wp.location) for k, wp in self.__waypoints.items()]

    def get_upcoming_waypoints(self):
        return [(k, wp.location) for k, wp in self.__waypoints.items() if not wp.is_visited()]

    def get_current_waypoint(self):
        wp = self.__waypoints[self.__current_waypoint_idx]
        return wp.location

    def get_move_action(self, state_tracker: StateTracker):
        # If we are done, don't do anything
        if self.is_done:
            return None

        # Check if the state tracker is for the right agent
        assert state_tracker.agent_id == self.__agent_id

        route = self.__get_route(state_tracker)

        agent_loc = state_tracker.get_memorized_state()[self.__agent_id]['location']

        if len(route) == 0 and self.is_done:
            return None
        elif agent_loc in route:
            move_action = route[agent_loc]
        else:
            agent_id = self.__agent_id
            raise Exception(f"Something went wrong with the path planning. The location {agent_loc} of agent "
                            f"{agent_id} is not found in the route.")

        return move_action

    def reset(self):
        self.is_done = False
        self.__current_waypoint_idx = 0
        for wp in self.__waypoints.values():
            wp.reset()

    def reset_full(self):
        # This function resets the navigator to a new instance
        self.__init__(self.__agent_id, self.__action_set, self.__algorithm, self.is_circular)

    def __get_current_waypoint(self):
        if self.__current_waypoint_idx is None:
            self.__current_waypoint_idx = 0

        wp = self.__waypoints[self.__current_waypoint_idx]

        return wp

    def __initialize_path_planner(self, algorithm, action_set):
        if algorithm == self.A_STAR_ALGORITHM:
            return AStarPlanner(action_set=action_set, metric=AStarPlanner.EUCLIDEAN_METRIC)
        else:
            raise Exception()

    def __update_waypoints(self, agent_loc):
        wp = self.__get_current_waypoint()
        if wp.is_visited(agent_loc):
            self.__current_waypoint_idx += 1

        if self.__current_waypoint_idx >= self.__nr_waypoints:
            self.is_done = True

    def __get_route(self, state_tracker: StateTracker):
        # Get our agent's location
        agent_loc = state_tracker.get_memorized_state()[state_tracker.agent_id]['location']

        # Update our waypoints based on agent's current location (if arrived at our current waypoint)
        self.__update_waypoints(agent_loc)

        # If we are done, we do nothing or start over if the path is circular
        if self.is_done:
            if self.is_circular:
                self.reset()
            else:
                return []

        # Get our occupation map
        self.__occupation_map, obj_grid = state_tracker.get_traversability_map(inverted=True)

        # Get our current waypoint
        current_wp = self.__get_current_waypoint()

        # Plan a path using the chosen path planning algorithm
        path = self.__path_planning_algo.plan(start=agent_loc, goal=current_wp.location,
                                              occupation_map=self.__occupation_map)

        # Go over the path and select the action that is required to go from one location to the other
        route = self.__get_route_from_path(agent_loc, path)

        return route

    def __get_route_from_path(self, agent_loc, path):
        route = {}
        curr_loc = agent_loc
        for idx, loc in enumerate(path):
            deltas = (loc[0] - curr_loc[0], loc[1] - curr_loc[1])
            if deltas not in self.__move_actions.values():
                agent_id = self.__agent_id
                raise Exception(f"Path leads from {curr_loc} to {loc} which requires a move of {deltas} which is not a "
                                f"possible move action for agent {agent_id}.")

            action = None
            for action_name, move_delta in self.__move_actions.items():
                if move_delta == deltas:
                    action = action_name

            route[curr_loc] = action

            curr_loc = loc

            if idx == len(path):
                route[loc] = None

        return route


class PathPlanner:

    def __init__(self, action_set):
        self.move_actions = get_move_actions(action_set)

    def plan(self, start, goal, occupation_map):
        pass


class AStarPlanner(PathPlanner):
    """ A* algorithm for path planning.
    """

    EUCLIDEAN_METRIC = "euclidean"
    MANHATTAN_METRIC = "manhattan"

    def __init__(self, action_set, metric=EUCLIDEAN_METRIC):
        super().__init__(action_set)
        if metric == self.EUCLIDEAN_METRIC:
            self.heuristic = lambda p1, p2: np.sqrt(np.sum((np.array(p1) - np.array(p2)) ** 2, axis=0))
        elif metric == self.MANHATTAN_METRIC:
            self.heuristic = lambda p1, p2: np.abs(p1[0] - p2[0]) + np.abs(p1[1] - p2[1])
        else:
            raise Exception(f"The distance metric {metric} for A* heuristic not known.")

    def plan(self, start, goal, occupation_map):
        """
        A star algorithm, returns the shortest path to get from goal to start.
        Uses an 2D numpy array, with 0 being traversable, anything else (e.g. 1) not traversable
        Implementation from:
        https://www.analytics-link.com/single-post/2018/09/14/Applying-the-A-Path-Finding-Algorithm-in-Python-Part-1-2D-square-grid
        """

        # possible movements
        neighbors = list(self.move_actions.values())

        close_set = set()
        came_from = {}
        gscore = {start: 0}
        fscore = {start: self.heuristic(start, goal)}
        oheap = []

        heapq.heappush(oheap, (fscore[start], start))

        while oheap:
            current = heapq.heappop(oheap)[1]

            if current == goal:
                path = []
                while current in came_from:
                    path.append(current)
                    current = came_from[current]
                return path[::-1]

            close_set.add(current)
            for i, j in neighbors:
                neighbor = current[0] + i, current[1] + j
                tentative_g_score = gscore[current] + self.heuristic(current, neighbor)
                if 0 <= neighbor[0] < occupation_map.shape[0]:
                    if 0 <= neighbor[1] < occupation_map.shape[1]:
                        if occupation_map[neighbor[0]][neighbor[1]] != 0:
                            continue
                    else:
                        # array bound y walls
                        continue
                else:
                    # array bound x walls
                    continue

                if neighbor in close_set and tentative_g_score >= gscore.get(neighbor, 0):
                    continue

                if tentative_g_score < gscore.get(neighbor, 0) or neighbor not in [i[1] for i in oheap]:
                    came_from[neighbor] = current
                    gscore[neighbor] = tentative_g_score
                    fscore[neighbor] = tentative_g_score + self.heuristic(neighbor, goal)
                    heapq.heappush(oheap, (fscore[neighbor], neighbor))

        # If no path is available we stay put
        return [start]


class Waypoint:

    def __init__(self, loc, priority):
        self.location = loc
        self.priority = priority
        self.__is_visited = False

    def is_visited(self, current_loc=None):
        if current_loc is None:
            return self.__is_visited
        elif current_loc[0] == self.location[0] and current_loc[1] == self.location[1] and not self.__is_visited:
            self.__is_visited = True

        return self.__is_visited

    def reset(self):
        self.__is_visited = False


def get_move_actions(action_set):
    move_actions = {}
    for action_name in action_set:
        if action_name == MoveNorth.__name__:
            move_actions[action_name] = (0, -1)
        elif action_name == MoveNorthEast.__name__:
            move_actions[action_name] = (1, -1)
        elif action_name == MoveEast.__name__:
            move_actions[action_name] = (1, 0)
        elif action_name == MoveSouthEast.__name__:
            move_actions[action_name] = (1, 1)
        elif action_name == MoveSouthWest.__name__:
            move_actions[action_name] = (-1, 1)
        elif action_name == MoveSouth.__name__:
            move_actions[action_name] = (0, 1)
        elif action_name == MoveWest.__name__:
            move_actions[action_name] = (-1, 0)
        elif action_name == MoveNorthWest.__name__:
            move_actions[action_name] = (-1, -1)

    # And moving nowhere is also possible
    move_actions[None] = (0, 0)

    return move_actions



import json
import random


class Message:
    """
    A simple object representing a communication message. An agent can create such a Message object by stating the
    content, its own id as the sender and (optional) a receiver. If a receiver is not given it is a message to all
    agents, including the sender.
    NOTE: this Message class is also used by the MATRX api

    Possible formats for mssg.to_id
    "agent1"                  = individual message to agent1 + message to team "agent1" if it exists
    ["agent1", "agent2"]      = 2 individual messages + 2 team messages if likewise named teams exist
    "team2"                   = team message. A team message is sent to everyone in that team
    '["agent3", "team4"]'     = team + agent message. Provided as a string via the api
    None                      = global message. This message is send to everyone
    """

    def __init__(self, content, from_id, to_id=None):
        """ Creates a message

        Parameters
        ----------
        content : anything
           Denotes the content of the message, can be anything: a string, dict, custom object, etc. As long
           as it is JSON serializable.
        from_id : str
           The ID who sent this message.
        to_id : str or list (optional, default, None)
           The ID to who to sent this message. If None, the message will be sent to all agents excluding the sender.
        """
        self.content = content  # content can be anything; a string, a dictionary, or even a custom object
        self.from_id = from_id  # the agent id who creates this message
        self.to_id = to_id  # the agent id who is the sender, when None it means all agents, including the sender
        self.message_id = self.__gen_random_string()  # randomly generated ID of the message

    def to_json(self):
        """ Make this class JSON serializable, such that it can be sent as JSON via the api """
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)

    def regen_id(self):
        """ Regenerates the message ID.
        When someone copies this message, by default the message_id is not changed. To avoid duplicate mssg IDs, this
        message can be called to change the regen the ID of a message.

        Examples
        ---------
        >>> print(mssg.message_id) # output: 242523098cb176b4
        >>> mssg.regen_id()
        >>> print(mssg.message_id) # output: 484de0a68fed7579
        """
        self.message_id = self.__gen_random_string()

    @staticmethod
    def __gen_random_string(length=32):
        """ Generates a random hexidecimal string of length 'length'.

        Parameters
        ----------
        length
            Length of the hexidecimal string to return.

        Returns
        -------
            A random hexidecimal string of length 'length'.
        """
        return '%030x' % random.randrange(16**length)



import copy
import itertools
import warnings

import numpy as np

import matrx
from matrx import utils
from matrx.objects.standard_objects import CollectionTarget, CollectionDropOffTile


class WorldGoal:
    """
    A class that tracks whether the simulation has reached its global goal.
    """

    def __init__(self):
        """
        We set the self.is_done to False as a start.
        """
        self.is_done = False

    def goal_reached(self, grid_world):
        """
        Returns whether the global goal of the simulated grid world is accomplished. This method should be overridden
        by a new goal function.

        Parameters
        ----------
        grid_world : GridWorld
            An up to date representation of the grid world that will be analyzed in this function on
            whether a specific coded global goal is reached.

        Returns
        -------
        goal_reached : bool
            True when the goal is reached, False otherwise.
        """
        pass

    def get_progress(self, grid_world):
        """
        Returns the progress of reaching the global goal in the simulated grid world. This method can be overridden
        if you want to track the progress. But is not required.

        Parameters
        ----------
        grid_world : GridWorld
            An up to date representation of the grid world that will be analyzed in this function on
            how far we are in obtaining the global world goal.

        Returns
        -------
        progress : float
            Representing with 0.0 no progress made, and 1.0 that the goal is reached.
        """
        pass

    def reset(self):
        """ Resets this goal's completion boolean and returns a copy of this object."""
        self.is_done = False
        return copy.deepcopy(self)


class LimitedTimeGoal(WorldGoal):
    """
    A world goal that simply tracks whether a maximum number of ticks has been reached.
    """

    def __init__(self, max_nr_ticks):
        """ Initialize the LimitedTimeGoal by saving the `max_nr_ticks`.
        """
        super().__init__()
        self.max_nr_ticks = max_nr_ticks

    def goal_reached(self, grid_world):
        """ Returns whether the number of specified ticks has been reached.

        Parameters
        ----------
        grid_world : GridWorld
            An up to date representation of the grid world that will be analyzed in this function on
            whether a specific coded global goal is reached.

        Returns
        -------
        goal_reached : bool
            True when the goal is reached, False otherwise.

        Examples
        --------

        For an example, see :meth:`matrx.grid_world.__check_simulation_goal`

        Checking all simulation goals from e.g. action, world goal, or somewhere else with access to the Gridworld,
        the function can be used as below:

        >>> goal_status = {}
        >>> if grid_world.simulation_goal is not None:
        >>>     if isinstance(grid_world.simulation_goal, list):
        >>>         for sim_goal in grid_world.simulation_goal:
        >>>             is_done = sim_goal.goal_reached(grid_world)
        >>>             goal_status[sim_goal] = is_done
        >>>     else:
        >>>         is_done = grid_world.simulation_goal.goal_reached(grid_world)
        >>>         goal_status[grid_world.simulation_goal] = is_done
        >>>
        >>> is_done = np.array(list(goal_status.values())).all()

        """
        nr_ticks = grid_world.current_nr_ticks
        if self.max_nr_ticks == np.inf or self.max_nr_ticks <= 0:
            self.is_done = False
        else:
            if nr_ticks >= self.max_nr_ticks:
                self.is_done = True
            else:
                self.is_done = False
        return self.is_done

    def get_progress(self, grid_world):
        """ Returns the progress of reaching the LimitedTimeGoal in the simulated grid world.

        Parameters
        ----------
        grid_world : GridWorld
            An up to date representation of the grid world that will be analyzed in this function on
            how far we are in obtaining the global world goal.

        Returns
        -------
        progress : float
            Representing with 0.0 no progress made, and 1.0 that the goal is reached.


        Examples
        --------
        Checking all simulation goals from e.g. action, world goal, or somewhere else with access to the Gridworld,
        the function can be used as below.
        In this example we know there is only 1 simulation goal.

        >>> progress = grid_world.simulation_goal.get_progress(grid_world)
        >>> print(f"The simulation is {progress * 100} percent complete!")


        """
        if self.max_nr_ticks == np.inf or self.max_nr_ticks <= 0:
            return 0.
        return min(1.0, grid_world.current_nr_ticks / self.max_nr_ticks)


class CollectionGoal(WorldGoal):

    def __init__(self, name, target_name, in_order=False):
        super().__init__()
        # Store the attributes
        self.__area_name = name
        self.__target_name = target_name
        self.__in_order = in_order

        # Set attributes we will use to speed up things and keep track of collected objects
        self.__drop_off_locs = None  # all locations where objects can be dropped off
        self.__target = None  # all (ordered) objects that need to be collected described in their properties
        self.__dropped_objects = {}  # a dictionary of the required dropped objects (id as key, tick as value)
        self.__attained_rank = 0  # The maximum attained rank of the correctly collected objects (only used if in_order)

    def goal_reached(self, grid_world):
        if self.__drop_off_locs is None:  # find all drop off locations, its tile ID's and goal blocks
            self.__drop_off_locs = []
            self.__find_drop_off_locations(grid_world)
            # Raise exception if no drop off locations were found.
            if len(self.__drop_off_locs) == 0:
                raise ValueError(f"The CollectionGoal {self.__area_name} could not find a "
                                 f"{CollectionDropOffTile.__name__} with its 'collection_area_name' set to "
                                 f"{self.__area_name}.")

        if self.__target is None:  # find all objects that need to be collected (potentially in order)
            self.__target = []
            self.__find_collection_objects(grid_world)

        # Go all drop locations and check if the requested objects are there (potentially dropped in the right order)
        is_satisfied = self.__check_completion(grid_world)
        self.is_done = is_satisfied

        return is_satisfied

    def __find_drop_off_locations(self, grid_world):
        all_objs = grid_world.environment_objects
        for obj_id, obj in all_objs.items():
            if 'name' in obj.properties.keys() \
                    and self.__area_name == obj.properties['name']:
                loc = obj.location
                self.__drop_off_locs.append(loc)

    def __find_collection_objects(self, grid_world):
        all_objs = grid_world.environment_objects
        for obj_id, obj in all_objs.items():
            if 'collection_zone_name' in obj.properties.keys() \
                    and self.__area_name == obj.properties['collection_zone_name']\
                    and 'collection_objects' in obj.properties and 'is_drop_off_target' in obj.properties\
                    and obj.properties['is_drop_off_target']:
                self.__target = obj.properties['collection_objects'].copy()

        # Raise warning if no target object was found.
        if len(self.__target) == 0:
            warnings.warn(f"The CollectionGoal {self.__area_name} could not find a {CollectionTarget.__name__} "
                          f"object or its 'collection_objects' property is empty.")

    def __check_completion(self, grid_world):
        # If we were already done before, we return the past values
        if self.is_done:
            return self.is_done

        # Get the current tick number
        curr_tick = grid_world.current_nr_ticks

        # Retrieve all objects and the drop locations (this is the most performance heavy; it loops over all drop locs
        # and queries the world to locate all objects at that point through distance calculation. Note: this calculation
        # is not required, as the range is zero!).
        obj_ids = [obj_id for loc in self.__drop_off_locs
                   for obj_id in grid_world.get_objects_in_range(loc, sense_range=0, object_type=None).keys()]

        # Get all world objects and agents
        all_objs = grid_world.environment_objects
        all_agents = grid_world.registered_agents
        all_ = {**all_objs, **all_agents}

        # Go through all objects at the drop off locations. If an object was not already detected before as a
        # required object, check if it is one of the desired objects. Also, ignore all drop off tiles and targets.
        detected_objs = {}
        for obj_id in obj_ids:
            obj_props = all_[obj_id].properties
            # Check if the object is either a collection area tile or a collection target object, if so skip it
            if ("is_drop_off" in obj_props.keys() and "collection_area_name" in obj_props.keys()) \
                    or ("is_drop_off_target" in obj_props.keys() and "collection_zone_name" in obj_props.keys()
                        and "is_invisible" in obj_props.keys()):
                continue
            for req_props in self.__target:
                obj_props = utils.__flatten_dict(obj_props)
                if req_props.items() <= obj_props.items():
                    detected_objs[obj_id] = curr_tick

        # Now compare the detected objects with the previous detected objects to see if any new objects were detected
        # and thus should be added to the dropped objects
        is_updated = False
        for obj_id in detected_objs.keys():
            if obj_id not in self.__dropped_objects.keys():
                is_updated = True
                self.__dropped_objects[obj_id] = detected_objs[obj_id]

        # Check if any objects detected previously are now not detected anymore, as such they need to be removed.
        removed = []
        for obj_id in self.__dropped_objects.keys():
            if obj_id not in detected_objs.keys():
                removed.append(obj_id)
        for obj_id in removed:
            is_updated = True
            self.__dropped_objects.pop(obj_id, None)

        # If required (and needed), check if the dropped objects are dropped in order by tracking the rank up which the
        # dropped objects satisfy the requested order.
        if self.__in_order and is_updated:
            # Sort the dropped objects based on the tick they were detected (in ascending order)
            sorted_dropped_obj = sorted(self.__dropped_objects.items(), key=lambda x: x[1], reverse=False)
            rank = 0
            for obj_id, tick in sorted_dropped_obj:
                props = all_[obj_id].properties
                props = utils.__flatten_dict(props)
                req_props = self.__target[rank]
                if req_props.items() <= props.items():
                    rank += 1
                else:
                    # as soon as the next object is not the one we expect, we stop the search at this attained rank.
                    break

            # The goal is done as soon as the attained rank is equal to the number of requested objects
            is_satisfied = rank == len(self.__target)
            # Store the attained rank, used to measure the progress
            self.__attained_rank = rank

        # objects do not need to be collected in order and new ones were dropped
        elif is_updated:
            # The goal is done when the number of collected objects equal the number of requested objects
            is_satisfied = len(self.__dropped_objects) == len(self.__target)

        # no new objects detected, so just return the past values
        else:
            is_satisfied = self.is_done

        return is_satisfied

    def get_progress(self, grid_world):
        # If we are done, just return 1.0
        if self.is_done:
            return 1.0

        # Check if the order matters, if so calculated the progress based on the maximum attained rank of correct
        # ordered collected objects.
        if self.__in_order:
            # Progress is the currently attained rank divided by the number of requested objects
            progress = self.__attained_rank / len(self.__target)

        # If the order does not matter, just calculate the progress as the number of correctly collected/dropped
        # objects.
        else:
            # Progress the is the number of collected objects divided by the total number of requested objects
            progress = len(self.__dropped_objects) / len(self.__target)

        return progress

    @classmethod
    def get_random_order_property(cls, possibilities, length=None, with_duplicates=False):
        """ Creates a `RandomProperty` representing a list of potential objects to collect in a certain order.

        Parameters
        ----------
        possibilities: iterable
            An iterable (e.g. list, tuple, etc.) of dictionaries representing property_name, property_value pairs that
            can be collected.

        length: int (optional, default None)
            The number of objects that need to be sampled from `possibilities` to be collected.

        with_duplicates: bool (optional, default False)
            Whether entries in `possibilities` can occur more than once in the lists.

        Returns
        -------
        RandomProperty
            A random property representing all possible lists of collection objects. Each list differs in the order of
            the objects. If length < len(possibilities), not all objects may be in each list. If with_duplicates=True,
            some objects might occur more than once in a list. This random property can be given to a `CollectionGoal`
            who will sample one of these lists every time a world is run. This allows a world with a `CollectionGoal`
            to denote different collection goals each time but still based on all properties in `possibilities`.

        Examples
        --------
        >>> from matrx import WorldBuilder
        >>> from matrx.logger import LogActions
        >>> from matrx.objects import SquareBlock
        >>> from matrx.goals import CollectionGoal
        >>> builder = WorldBuilder(shape=(3, 3))
        >>> builder.add_object([0, 0], "Red Block", callable_class=SquareBlock, visualize_colour="#eb4034")
        >>> builder.add_object([1, 1], "Blue Block", callable_class=SquareBlock, visualize_colour="#3385e8")

        Add a collection goal, where we should collect red and blue blocks but every time we run the world, in a different
         order. To do so, we need to pass a RandomProperty to `add_collection_goal` which it uses to sample such an
         order each created world. We call this utility method to get us such a RandomProperty.
        >>> rp_order = CollectionGoal.get_random_order_property([{'visualize_colour': 'eb4034'}, {'visualize_colour': '3385e8'}])
        >>> builder.add_collection_goal("Drop", [(2, 2)], rp_order, in_order=True)

        See Also
        --------
        :any:`~matrx.WorldBuilder.add_collection_goal`
            The method that receives this return value.
        :any:`~matrx.WorldBuilder.RandomProperty`
            The class representing a property with a random value each world creation.

        """
        if length is None:
            length = len(possibilities)

        if not with_duplicates:
            orders = itertools.permutations(possibilities, r=length)
        else:  # with_duplicates
            orders = itertools.product(possibilities, repeat=length)
        orders = list(orders)

        rp_orders = matrx.world_builder.RandomProperty(values=orders)

        return rp_orders



from matrx.agents import AgentBrain


class BlockWorldAgent(AgentBrain):

    def __init__(self, memorize_for_ticks=10):
        self.__memorize_for_ticks = memorize_for_ticks
        self.__collect = None
        super().__init__()

    def initialize(self):
        pass

    def filter_observations(self, state):
        # self.state[self.agent_id]
        # self.state["foo"]
        # self.state[["Collect block_585", "Collect block_586"]]
        # self.state[{"foo": "bar"}]
        # self.state["is_open"]
        # self.state[{"room_name": "room_0"}]
        # self.state[{"room_name": "room_0", "class_inheritance": "Wall"}]
        # self.state[{"room_name": ["room_0", "room_1"]}]
        # self.state[{"room_name": ["room_0", "room_1"], "class_inheritance": "Wall"}]
        # self.state[{"room_name": ["room_0", "room_1"], "class_inheritance": ["Wall", "Door"]}]
        # self.state[{"room_name": ["room_0", "room_1"], "is_open": True}]
        #
        # self.state.get_world_info()
        # self.state.get_agents()
        # for n in self.state.get_all_room_names():
        #     self.state.get_room_doors(room_name=n)
        #     self.state.get_room_objects(room_name=n)
        #     self.state.get_room(room_name=n)
        # self.state.get_agents_with_property("team")
        # self.state.get_closest_agents()
        # self.state.get_closest_objects()
        # self.state.get_closest_room_door()
        # self.state.get_closest_with_property("room_name", "Drop_off")
        # self.state.get_distance_map()
        # self.state.get_of_type('Wall')
        # self.state.get_team_members("")
        # self.state.get_traverse_map()
        # self.state.remove_with_property({"room_name": "world_bounds"})

        if self.__collect is None:
            pass
        return self.state

    def decide_on_action(self, state_dict):
        return None, {}



import threading
import logging
from flask import Flask, render_template, request, jsonify, send_from_directory

'''
This file holds the code for the MATRX RESTful api. 
External scripts can send POST and/or GET requests to retrieve state, tick and other information, and send 
userinput or other information to MATRX. The api is a Flask (Python) webserver.

For visualization, see the seperate MATRX visualization folder / package.
'''

debug = True
port = 3000
app = Flask(__name__, template_folder='templates')

# the path to the media folder of the user (outside of the MATRX package)
ext_media_folder = ""

#########################################################################
# Visualization server routes
#########################################################################

@app.route('/human-agent/<id>')
def human_agent_view(id):
    """
    Route for HumanAgentBrain

    Parameters
    ----------
    id
        The human agent ID. Is obtained from the URL.

    Returns
    -------
    str
        The template for this agent's view.

    """
    return render_template('human_agent.html', id=id)


# route for agent, get the ID from the URL
@app.route('/agent/<id>')
def agent_view(id):
    """
    Route for AgentBrain

    Parameters
    ----------
    id
        The agent ID. Is obtained from the URL.

    Returns
    -------
    str
        The template for this agent's view.

    """
    return render_template('agent.html', id=id)


@app.route('/god')
def god_view():
    """
    Route for the 'god' view which contains the ground truth of the world without restrictions.

    Returns
    -------
    str
        The template for this view.

    """
    return render_template('god.html')


@app.route('/')
@app.route('/start')
def start_view():
    """
    Route for the 'start' view which shows information about the current scenario, including links to all agents.

    Returns
    -------
    str
        The template for this view.

    """
    return render_template('start.html')




@app.route('/shutdown_visualizer', methods=['GET', 'POST'])
def shutdown():
    """ Shuts down the visualizer by stopping the Flask thread

    Returns
        True
    -------
    """
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Unable to shutdown visualizer server. Not running with the Werkzeug Server')
    func()
    print("Visualizer server shutting down...")
    return jsonify(True)


@app.route('/fetch_external_media/<path:filename>')
def external_media(filename):
    """ Facilitate the use of images in the visualization outside of the static folder

    Parameters
    ----------
    filename
        path to the image file in the external media folder of the user.

    Returns
    -------
        Returns the url (relative from the website root) to that file
    """
    return send_from_directory(ext_media_folder, filename, as_attachment=True)


#########################################################################
# Visualization Flask methods
#########################################################################

def flask_thread():
    """
    Starts the Flask server on localhost:3000
    """

    if not debug:
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    app.run(host='0.0.0.0', port=port, debug=False, use_reloader=False)

def run_matrx_visualizer(verbose, media_folder):
    """
    Creates a seperate Python thread in which the visualization server (Flask) is started, serving the JS visualization
    :return: MATRX visualization Python thread
    """
    global debug, ext_media_folder
    debug = verbose
    ext_media_folder = media_folder

    print("Starting visualization server")
    print("Initialized app:", app)
    vis_thread = threading.Thread(target=flask_thread)
    vis_thread.start()
    return vis_thread

if __name__ == "__main__":
    run_matrx_visualizer()


import warnings
import copy

from matrx.actions.object_actions import GrabObject, DropObject, RemoveObject
from matrx.actions.door_actions import OpenDoorAction, CloseDoorAction
from matrx.agents.agent_utils.state import State
from matrx.agents.agent_utils.state_tracker import StateTracker
from matrx.agents.agent_brain import AgentBrain
import numpy as np

from matrx.messages import Message


class HumanAgentBrain(AgentBrain):
    """ Creates an Human Agent which is an agent that can be controlled by a human.
    """

    def __init__(self, memorize_for_ticks=None, fov_occlusion=False, max_carry_objects=3,
                 grab_range=1, drop_range=1, door_range=1, remove_range=1):
        """ Creates an Human Agent which is an agent that can be controlled by a human.
        """
        super().__init__(memorize_for_ticks=memorize_for_ticks)
        self.__fov_occlusion = fov_occlusion
        if fov_occlusion:
            warnings.warn("FOV Occlusion is not yet fully implemented. Setting fov_occlusion to True has not effect.")
        self.__max_carry_objects = max_carry_objects
        self.__remove_range = remove_range
        self.__grab_range = grab_range
        self.__drop_range = drop_range
        self.__door_range = door_range
        self.__remove_range = remove_range

    def _factory_initialise(self, agent_name, agent_id, action_set, sense_capability, agent_properties,
                            customizable_properties, rnd_seed, callback_is_action_possible, key_action_map=None):
        """ Called by the WorldFactory to initialise this agent with all required properties in addition with any custom
        properties.
        This also sets the random number generator with a seed generated based on the random seed of the
        world that is generated.

        Note; This method should NOT be overridden!

        Parameters
        ----------
        agent_name : str
            The name of the agent.
        agent_id: str
            The unique ID given by the world to this agent's avatar. So the agent knows what body is his.
        action_set : List
            The list of action names this agent is allowed to perform.
        sense_capability : SenseCapability
            The SenseCapability of the agent denoting what it can see withing what range.
        agent_properties : dict
            The dictionary of properties containing all mandatory and custom properties.
        customizable_properties : list
            A list of keys in agent_properties that this agent is allowed to change.
        rnd_seed : int
            The random seed used to set the random number generator self.rng
        key_action_map : (optional, default, None)
            Maps user pressed keys (e.g. arrow key up) to a specific action. See this link for the available keys
            https://developer.mozilla.org/nl/docs/Web/API/KeyboardEvent/key/Key_Values
        """

        # The name of the agent with which it is also known in the world
        self.agent_name = agent_name

        # The id of the agent
        self.agent_id = agent_id

        # The names of the actions this agent is allowed to perform
        self.action_set = action_set

        # Setting the random seed and rng
        self.rnd_seed = rnd_seed
        self._set_rnd_seed(seed=rnd_seed)

        # The SenseCapability of the agent; what it can see and within what range
        self.sense_capability = sense_capability

        # Contains the agent_properties
        self.agent_properties = agent_properties

        # Specifies the keys of properties in self.agent_properties which can  be changed by this Agent in this file. If
        # it is not writable, it can only be  updated through performing an action which updates that property (done by
        # the environment).
        # NOTE: Changing which properties are writable cannot be done during runtime! Only in  the scenario manager
        self.keys_of_agent_writable_props = customizable_properties

        # A callback to the GridWorld instance that can check whether any action (with its arguments) will succeed and
        # if not why not (in the form of an ActionResult).
        self.__callback_is_action_possible = callback_is_action_possible

        # a list which maps user inputs to actions, defined in the scenario manager
        if key_action_map is None:
            self.key_action_map = {}
        else:
            self.key_action_map = key_action_map

        # Initializing the State object
        self._init_state()

    def _get_action(self, state, agent_properties, agent_id, user_input):
        """
        The function the environment calls. The environment receives this function object and calls it when it is time
        for this agent to select an action.

        The function overwrites the default get_action() function for normal agents,
        and instead executes the action commanded by the user, which is received
        via the api from e.g. a visualization interface.

        Note; This method should NOT be overridden!

        Parameters
        ----------
        state : dict
            A state description containing all properties of EnvObject that are within a certain range as defined by
            self.sense_capability. It is a list of properties in a dictionary
        agent_properties : dict
            The properties of the agent, which might have been changed by the environment as a result of actions of
            this or other agents.
        agent_id : str
            the ID of this agent
        user_input : list
            any user input given by the user for this human agent via the api

        Returns
        -------
         filtered_state : dict
            The filtered state of this agent
        agent_properties : dict
            the agent properties which the agent might have changed,
        action : str
            an action string, which is the class name of one of the actions in the Action package.
        action_kwargs : dict
            Keyword arguments for the action

        """
        # Process any properties of this agent which were updated in the environment as a result of
        # actions
        self.agent_properties = agent_properties

        # Update the state property of an agent with the GridWorld's state dictionary
        self._state.state_update(state)

        # Call the filter method to filter the observation
        self._state = self.filter_observations(self._state)
        if isinstance(self._state, dict):
            raise ValueError(f"The filter_observation function of "
                             f"{self.agent_id} does not return a State "
                             f"object, but a dictionary. Please return "
                             f"self.state.")

        # only keep user input which is actually connected to an agent action
        usrinput = self.filter_user_input(user_input)

        # Call the method that decides on an action
        action, action_kwargs = self.decide_on_action(self._state, usrinput)

        # Store the action so in the next call the agent still knows what it did
        self.previous_action = action

        # Get the dictionary from the State object
        filtered_state = self._state.as_dict()

        # Return the filtered state, the (updated) properties, the intended actions and any keyword arguments for that
        # action if needed.
        return filtered_state, self.agent_properties, action, action_kwargs

    def decide_on_action(self, state, user_input):
        """ Contains the decision logic of the agent.

        This method determines what action the human agent will perform. The GridWorld is responsible for deciding when
        an agent can perform an action again, if so this method is called for each agent. Two things need to be
        determined, which action and with what arguments.

        The action is returned simply as the class name (as a string), and the action arguments as a dictionary with the
        keys the names of the keyword arguments. An argument that is always possible is that of action_duration, which
        denotes how many ticks this action should take (e.g. a duration of 1, makes sure the agent has to wait 1 tick).

        Note; this function of the human_agent_brain overwrites the decide_on_action() function of the default agent,
        also providing the user input.


        Parameters
        ==========
        state : dict
            A state description containing all properties of EnvObject that are within a certain range as
            defined by self.sense_capability. It is a list of properties in a dictionary

        user_input : list
            A dictionary containing the key presses of the user, intended for controlling thus human agent.

        Returns
        =======
        action_name : str
            A string of the class name of an action that is also in self.action_set. To ensure backwards compatibility
            you could use Action.__name__ where Action is the intended action.

        action_args : dict
            A dictionary with keys any action arguments and as values the actual argument values. If a required argument
            is missing an exception is raised, if an argument that is not used by that action a warning is printed. The
            argument applicable to all action is `action_duration`, which sets the number ticks the agent is put on hold
            by the GridWorld until the action's world mutation is actual performed and the agent can perform a new
            action (a value of 0 is no wait, 1 means to wait 1 tick, etc.).
        """

        action_kwargs = {}

        # if no keys were pressed, do nothing
        if user_input is None or user_input == []:
            return None, {}

        # take the latest pressed key (for now), and fetch the action associated with that key
        pressed_keys = user_input[-1]
        action = self.key_action_map[pressed_keys]

        # if the user chose a grab action, choose an object within a grab_range of 1
        if action == GrabObject.__name__:
            # Assign it to the arguments list
            action_kwargs['grab_range'] = self.__grab_range  # Set grab range
            action_kwargs['max_objects'] = self.__max_carry_objects  # Set max amount of objects
            action_kwargs['object_id'] = None

            obj_id = self.__select_random_obj_in_range(state, range_=self.__grab_range, property_to_check="is_movable")
            action_kwargs['object_id'] = obj_id

        # If the user chose to drop an object in its inventory
        elif action == DropObject.__name__:
            # Assign it to the arguments list
            action_kwargs['drop_range'] = self.__drop_range  # Set drop range

        # If the user chose to remove an object
        elif action == RemoveObject.__name__:
            # Assign it to the arguments list
            action_kwargs['remove_range'] = self.__remove_range  # Set drop range

            obj_id = self.__select_random_obj_in_range(state, range_=self.__remove_range, property_to_check="is_movable")
            action_kwargs['object_id'] = obj_id

        # if the user chose to do an open or close door action, find a door to open/close within range
        elif action == OpenDoorAction.__name__ or action == CloseDoorAction.__name__:
            action_kwargs['door_range'] = self.__door_range
            action_kwargs['object_id'] = None

            # Get all doors from the perceived objects
            objects = list(state.keys())
            doors = [obj for obj in objects if 'is_open' in state[obj]]

            # get all doors within range
            doors_in_range = []
            for object_id in doors:
                # Select range as just enough to grab that object
                dist = int(np.ceil(np.linalg.norm(
                    np.array(state[object_id]['location']) - np.array(
                        state[self.agent_id]['location']))))
                if dist <= action_kwargs['door_range']:
                    doors_in_range.append(object_id)

            # choose a random door within range
            if len(doors_in_range) > 0:
                action_kwargs['object_id'] = self.rnd_gen.choice(doors_in_range)

        return action, action_kwargs

    def filter_observations(self, state):
        """
        All our agent work through the OODA-loop paradigm; first you observe, then you orient/pre-process, followed by
        a decision process of an action after which we act upon the action.

        However, as a human agent is controlled by a human, only the observe part is executed.

        This is the Observe phase. In this phase you filter the state further to only those properties the agent is
        actually SUPPOSED to see. Since the grid world returns ALL properties of ALL objects within a certain range(s),
        but perhaps some objects are obscured because they are behind walls, or an agent is not able to see some
        properties an certain objects.

        This filtering is what you do here.

        Parameters
        ----------
        state : dict
            A state description containing all properties of EnvObject that are within a certain range as
            defined by self.sense_capability. It is a list of properties in a dictionary

        Returns
        -------
         filtered_state : dict
            The filtered state of this agent

        """
        return state

    def filter_user_input(self, user_input):
        """ From the received userinput, only keep those which are actually connected to a specific agent action.

        Parameters
        ----------
        user_input : list
            A dictionary containing the key presses of the user, intended for controlling thus human agent.

        """

        # read messages and remove them
        for message in list(self.received_messages):
            print("Received message:", message)
            self.received_messages.remove(message)

        if user_input is None:
            return []
        possible_key_presses = list(self.key_action_map.keys())
        return list(set(possible_key_presses) & set(user_input))

    def create_context_menu_for_self(self, clicked_object_id, click_location, self_selected):
        """ Generate options for a context menu for a specific object/location which the user controlling this
        human agent opened.

        For the default MATRX visualization, the context menu is opened by right clicking on an object. This function
        should generate a list of options (actions, messages, or something else) which relate to that object.
        Each option is in the shape of a text shown in the context menu, and a message which is send to this agent if
        the user actually clicks that context menu option.

        Parameters
        ----------
        clicked_object_id : str
            A string indicating the ID of an object. Is None if the user clicked on a background tile (which has no ID).
        click_location : list
            A list containing the [x,y] coordinates of the object on which the user right clicked.
        self_selected : bool
            Describes if the current human agent being controlled by the user was selected or not before opening the
            context menu. Depending on this, you might pass back a different context menu in this function.
            E.g. option 1: no-one selected + right click is the same as self selected + right click: both open the
            current agent's context menu.
            option 2: self selected + right click opens our own context menu, no one selected + right click gives a
            context menu with commands for the entire TEAM.

        Returns
        -------
         context_menu : list
            A list containing context menu items. Each context menu item is a dict with a 'OptionText' key, which is
            the text shown in the menu for the option, and a 'Message' key, which is the message instance that is sent
            to this agent when the user clicks on the context menu option.
        """
        print("Context menu self with self selected:", self_selected)

        context_menu = []

        for action in self.action_set:

            context_menu.append({
                "OptionText": f"Do action: {action}",
                "Message": Message(content=action, from_id=self.agent_id, to_id=self.agent_id)
            })

        return context_menu


    def create_context_menu_for_other(self, agent_id_who_clicked, clicked_object_id, click_location):
        """ Generate options for a context menu for a specific object/location that a user NOT controlling this
        human agent opened.

        Thus: another human agent selected this agent, opened a context menu by right clicking on an object or location.
        This function is called. It should return actions, messages, or other info for what this agent can do for that
        object / location.

        Example usecase: tasking another agent that is not yourself, e.g. to move an object.

        For the default MATRX visualization, the context menu is opened by right clicking on an object. This function
        should generate a list of options (actions, messages, or something else) which relate to that object or location.
        Each option is in the shape of a text shown in the context menu, and a message which is send to this agent if
        the user actually clicks that context menu option.

        Parameters
        ----------
        agent_id_who_clicked : str
            The ID of the (human) agent that selected this agent and requested for a context menu.
        clicked_object_id : str
            A string indicating the ID of an object. Is None if the user clicked on a background tile (which has no ID).
        click_location : list
            A list containing the [x,y] coordinates of the object on which the user right clicked.

        Returns
        -------
         context_menu : list
            A list containing context menu items. Each context menu item is a dict with a 'OptionText' key, which is
            the text shown in the menu for the option, and a 'Message' key, which is the message instance that is sent
            to this agent when the user clicks on the context menu option.
        """
        print("Context menu other")
        context_menu = []

        # Generate a context menu option for every action
        for action in self.action_set:
            context_menu.append({
                "OptionText": f"Do action: {action}",
                "Message": Message(content=action, from_id=clicked_object_id, to_id=self.agent_id)
            })
        return context_menu



    def __select_random_obj_in_range(self, state, range_, property_to_check=None):

        # Get all perceived objects
        object_ids = list(state.keys())

        # Remove world from state
        object_ids.remove("World")

        # Remove self
        object_ids.remove(self.agent_id)

        # Remove all (human)agents
        object_ids = [obj_id for obj_id in object_ids if "AgentBrain" not in state[obj_id]['class_inheritance'] and
                      "AgentBody" not in state[obj_id]['class_inheritance']]

        # find objects in range
        object_in_range = []
        for object_id in object_ids:

            # Select range as just enough to grab that object
            dist = int(np.ceil(np.linalg.norm(np.array(state[object_id]['location'])
                                              - np.array(state[self.agent_id]['location']))))
            if dist <= range_:
                # check for any properties specifically specified by the user
                if property_to_check is not None:
                    if property_to_check in state[object_id] and state[object_id][property_to_check]:
                        object_in_range.append(object_id)
                else:
                    object_in_range.append(object_id)

        # Select an object if there are any in range
        if object_in_range:
            object_id = self.rnd_gen.choice(object_in_range)
        else:
            object_id = None

        return object_id





import os

from matrx.agents.agent_types.human_agent import HumanAgentBrain
from matrx.agents.agent_types.patrolling_agent import PatrollingAgentBrain
from matrx.logger.log_agent_actions import LogActions
from matrx.world_builder import WorldBuilder
from matrx.actions.move_actions import *


def create_builder():
    tick_dur = 0.01
    factory = WorldBuilder(random_seed=1, shape=[15, 6], tick_duration=tick_dur, verbose=False,
                           simulation_goal=int(300/tick_dur),
                           run_matrx_api=True, run_matrx_visualizer=True, visualization_bg_clr="#000000",
                           visualization_bg_img="/static/images/soesterberg_luchtfoto.jpg")

    factory.add_logger(logger_class=LogActions, save_path="log_data/")

    even = True
    for x in range(15):
        waypoints = [(x, 0), (x, 5)]
        navigating_agent = PatrollingAgentBrain(waypoints)
        human_agent = HumanAgentBrain()
        if even:
            even = False
            start = [x, 0]
            factory.add_agent(start, navigating_agent, name="Navigate_" + str(x), visualize_shape=0, has_menu=True)
        else:
            even = True
            start = [x, 5]
            key_action_map = {
                'w': MoveNorth.__name__,
                'd': MoveEast.__name__,
                's': MoveSouth.__name__,
                'a': MoveWest.__name__
            }
            factory.add_human_agent(start, human_agent, name="human_" + str(x),
                                    key_action_map=key_action_map, visualize_shape='img',
                                    img_name="/static/images/transparent.png")

    factory.add_line(start=[1, 1], end=[3, 1], name="T")
    factory.add_line(start=[2, 2], end=[2, 4], name="T")

    factory.add_line(start=[5, 1], end=[5, 4], name="N")
    factory.add_line(start=[6, 2], end=[7, 3], name="N")
    factory.add_line(start=[8, 1], end=[8, 4], name="N")

    factory.add_line(start=[11, 1], end=[12, 1], name="O")
    factory.add_line(start=[10, 2], end=[10, 3], name="O")
    factory.add_line(start=[11, 4], end=[12, 4], name="O")
    factory.add_line(start=[13, 2], end=[13, 3], name="O")
    factory.add_object((4, 3), "Object", visualize_shape='img', img_name="/static/images/fire.gif")

    factory.add_smoke_area([0, 0], width=15, height=6, name="smoke", smoke_thickness_multiplier=0.5)
    return factory


def run_test(nr_of_worlds=1):
    builder = create_builder()

    # startup world-overarching MATRX scripts, such as the api and/or visualizer if requested
    media_folder = os.path.dirname(os.path.realpath(__file__)) # set our path for media files to our current folder
    builder.startup(media_folder=media_folder)

    # run each world
    for world in builder.worlds(nr_of_worlds=nr_of_worlds):
        world.run(builder.api_info)

    # stop MATRX scripts such as the api and visualizer (if used)
    builder.stop()



# Aliasing imports
from matrx.world_builder import WorldBuilder


__docformat__ = "restructuredtext"


######
# We do this so we are sure everything is imported and thus can be found
# noinspection PyUnresolvedReferences
import pkgutil

__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    __all__.append(module_name)
    _module = loader.find_module(module_name).load_module(module_name)
    globals()[module_name] = _module
######

# Set package attributes
name = "MATRX: Man-Agent Teaming - Rapid Experimentation Software"
